#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import abc
import copy
import dataclasses
import itertools
import pathlib
import re
import shutil
from typing import Optional, Union, List

import numpy as np

from ase import Atoms
from ase.io import read, write
from ase.geometry import find_mic
from ase.constraints import FixAtoms
from ase.neb import interpolate, idpp_interpolate

from .. import parse_constraint_info
from ..reactor import AbstractReactor
from ..utils import plot_bands, plot_mep, compute_rxn_coords


@dataclasses.dataclass
class StringReactorSetting:

    #: Reactor setting.
    backend: str = "external"

    #: Period to save the trajectory.
    dump_period: int = 1

    #: Period to save the restart file.
    ckpt_period: int = 100

    #: Number of images along the pathway.
    nimages: int = 7

    #: Methods that interpolates structures.
    interpolation: dict = dataclasses.field(default_factory=dict)

    #: Optimiser.
    optimiser: str = "bfgs"

    #: Spring constant, eV/Ang2.
    k: float = 5.0

    #: Whether use CI-NEB.
    climb: bool = False

    #: Convergence coordinate RMS.
    rrms: float = 0.5

    #: Convergence coordinate MAX.
    rmax: float = 0.5

    #: Convergence force RMS.
    frms: float = 0.25

    #: Convergence force tolerance.
    fmax: float = 0.05

    #: Maximum number of steps.
    steps: int = 100

    #: FixAtoms.
    constraint: Optional[str] = None

    #: Parameters that are used to update.
    _internals: dict = dataclasses.field(default_factory=dict)

    def get_init_params(self):
        """"""

        return copy.deepcopy(self._internals)

    def get_run_params(self):
        """"""

        raise NotImplementedError(
            f"{self.__class__.__name__} has no function for run params."
        )


class AbstractStringReactor(AbstractReactor):

    name: str = "string"

    traj_name: str = "nebtraj.xyz"

    @AbstractReactor.directory.setter
    def directory(self, directory_: Union[str, pathlib.Path]):
        self._directory = pathlib.Path(directory_)
        # avoid inconsistent in ASE
        # the actual calc directory will be set in _irun
        self.calc.directory = str(self.directory)

        return

    def run(self, structures: List[Atoms], read_ckpt=True, *args, **kwargs):
        """"""
        super().run(structures=structures, *args, **kwargs)

        # - compatibility
        read_cache = kwargs.get("read_cache", None)
        if read_cache is not None:
            read_ckpt = read_cache

        # - Double-Ended Methods...
        ini_atoms, fin_atoms = structures
        try:
            self._print(f"ini_atoms: {ini_atoms.get_potential_energy()}")
            self._print(f"fin_atoms: {fin_atoms.get_potential_energy()}")
        except RuntimeError:
            # RuntimeError: Atoms object has no calculator.
            self._print("Not energies attached to IS and FS.")

        # - backup old parameters
        prev_params = copy.deepcopy(self.calc.parameters)

        # -
        if not self._verify_checkpoint():
            self._debug(f"... start from the scratch @ {self.directory.name} ...")
            self.directory.mkdir(parents=True, exist_ok=True)
            self._irun([ini_atoms, fin_atoms], *args, **kwargs)
        else:
            self._debug(f"... restart @ {self.directory.name} ...")
            # - check if converged
            converged = self.read_convergence()
            if not converged:
                self._debug(f"... unconverged @ {self.directory.name} ...")
                ckpt_wdir = self._save_checkpoint() if read_ckpt else None
                self._debug(f"... checkpoint @ {str(ckpt_wdir)} ...")
                self._irun(structures, ckpt_wdir=ckpt_wdir, *args, **kwargs)
            else:
                self._debug(f"... converged @ {self.directory.name} ...")

        self.calc.parameters = prev_params
        self.calc.reset()

        # - check again
        curr_band, converged = None, self.read_convergence()
        if converged:
            self._debug(f"... 2. converged @ {self.directory.name} ...")
            band_frames = self.read_trajectory()  # (nbands, nimages)
            if band_frames:
                # FIXME: make below a function
                plot_mep(self.directory, band_frames[-1])
                write(self.directory / "temptraj.xyz", itertools.chain(*band_frames))

                curr_band = band_frames[-1]

                rxn_coords = compute_rxn_coords(curr_band)

                energies = [a.get_potential_energy() for a in curr_band]
                imax = 1 + np.argsort(energies[1:-1])[-1]
                # NOTE: maxforce in cp2k is norm(atomic_forces)
                maxfrc = np.max(curr_band[imax].get_forces(apply_constraint=True))

                self._print(
                    f"rxncoords: {rxn_coords[0]:.2f} -> {rxn_coords[imax]:.2f} "
                    + f"-> {rxn_coords[-1]:.2f}"
                )
                self._print(
                    f"maxfrc: {maxfrc} Ea_f: {energies[imax]-energies[0]:<8.4f} "
                    + f"dE: {energies[-1]-energies[0]:<8.4f}"
                )
            else:
                self._debug(f"... CANNOT read bands @ {self.directory.name} ...")
                ...
        else:
            self._debug(f"... 2. unconverged @ {self.directory.name} ...")

        return curr_band

    @abc.abstractmethod
    def _irun(self, structures: List[Atoms], *args, **kwargs):
        """"""

        return

    def _verify_checkpoint(self, *args, **kwargs) -> bool:
        """Check if the current directory has any valid outputs or it just created
        the input files.

        """

        return self.directory.exists()

    def _save_checkpoint(self, *args, **kwargs):
        """"""
        # - find previous runs...
        prev_wdirs = sorted(self.directory.glob(r"[0-9][0-9][0-9][0-9][.]run"))
        self._debug(f"prev_wdirs: {prev_wdirs}")
        curr_index = len(prev_wdirs)

        curr_wdir = self.directory / f"{str(curr_index).zfill(4)}.run"
        self._debug(f"curr_wdir: {curr_wdir}")

        # - backup files
        curr_wdir.mkdir()
        for x in self.directory.iterdir():
            if not re.match(r"[0-9]{4}\.run", x.name):
                # NOTE: default is to move everything to the new folder
                # if x.name in self.saved_fnames:
                #    shutil.move(x, curr_wdir)
                # else:
                #    x.unlink()
                shutil.move(x, curr_wdir)
            else:
                ...

        return curr_wdir

    def _preprocess_constraints(self, atoms, cons_text: str) -> None:
        """"""
        atoms._del_constraints()
        mobile_indices, frozen_indices = parse_constraint_info(
            atoms, cons_text, ignore_ase_constraints=True, ret_text=False
        )
        if frozen_indices:
            atoms.set_constraint(FixAtoms(indices=frozen_indices))

        return

    def _align_structures(
        self, structures: List[Atoms], run_params: dict, *args, **kwargs
    ) -> List[Atoms]:
        """Create a reaction pathway based on two structures.

        Args:
            structures: Two structures - Initial state and Final State.
            run_params: We need thet latest `constraint` information.

        Returns:
            A List of Atoms structures.

        """
        nstructures = len(structures)
        if nstructures == 2:
            self._print("Interpolate a pathway.")
            # - check lattice consistency
            ini_atoms, fin_atoms = structures
            c1, c2 = ini_atoms.get_cell(complete=True), fin_atoms.get_cell(
                complete=True
            )
            assert np.allclose(c1, c2), "Inconsistent unit cell..."

            cons_text = run_params.pop("constraint", None)
            self._preprocess_constraints(ini_atoms, cons_text)
            self._preprocess_constraints(fin_atoms, cons_text)

            # - 
            use_mic = self.setting.interpolation.get("mic", True)
            idpp_params = self.setting.interpolation.get("idpp", {})

            # - linear interpolate
            shifts = fin_atoms.get_positions() - ini_atoms.get_positions()
            if use_mic:
                self._print("Align IS and FS based on MIC.")
                curr_vectors, curr_distances = find_mic(shifts, c1, pbc=True)
                # self._debug(f"curr_vectors: {curr_vectors}")
                self._print(f"disp: {np.linalg.norm(curr_vectors)}")
                fin_atoms.positions = ini_atoms.get_positions() + curr_vectors
            else:
                self._print(f"disp: {np.linalg.norm(shifts)}")

            nimages = self.setting.nimages
            images = [ini_atoms]
            images += [ini_atoms.copy() for i in range(nimages - 2)]
            images.append(fin_atoms)

            interpolate(
                images=images,
                mic=False,
                interpolate_cell=False,
                use_scaled_coord=False,
                apply_constraint=False,
            )

            if idpp_params:
                # FIXME: make idpp a manager?
                idpp_interpolate(
                    images=images, traj=str(self.directory/"idpp_images.traj"),
                    log=str(self.directory/"idpp.log"), mic=use_mic, **idpp_params
                )
        else:
            self._print("Use a pre-defined pathway.")
            images = [a.copy() for a in structures]

        return images

    def read_trajectory(self, *args, **kwargs):
        """"""
        # - find previous runs...
        prev_wdirs = sorted(self.directory.glob(r"[0-9][0-9][0-9][0-9][.]run"))
        self._debug(f"prev_wdirs: {prev_wdirs}")

        traj_list = []
        for w in prev_wdirs:
            curr_frames = self._read_a_single_trajectory(wdir=w)
            traj_list.append(curr_frames)

        cache_nebtraj = self.directory / self.traj_name
        if cache_nebtraj.exists() and cache_nebtraj.stat().st_size != 0:
            traj_list.append(self._read_a_single_trajectory(wdir=self.directory))

        # - concatenate
        traj_frames, ntrajs = [], len(traj_list)
        if ntrajs > 0:
            traj_frames.extend(traj_list[0])
            for i in range(1, ntrajs):
                prev_end_band, curr_beg_band = traj_list[i - 1][-1], traj_list[i][0]
                for j, (a, b) in enumerate(zip(prev_end_band, curr_beg_band)):
                    assert np.allclose(
                        a.positions, b.positions
                    ), f"Traj {i-1} and traj {i} are not consecutive in positions."
                traj_frames.extend(traj_list[i][1:])
        else:
            ...

        if traj_frames:
            # FIXME: make below a function
            plot_mep(self.directory, traj_frames[-1])

            curr_band = traj_frames[-1]

            rxn_coords = compute_rxn_coords(curr_band)

            energies = [a.get_potential_energy() for a in curr_band]
            imax = 1 + np.argsort(energies[1:-1])[-1]
            # NOTE: maxforce in cp2k is norm(atomic_forces)
            maxfrc = np.max(curr_band[imax].get_forces(apply_constraint=True))

            self._print(f"imax: {imax}")
            self._print(
                f"rxncoords: {rxn_coords[0]:.2f} -> {rxn_coords[imax]:.2f} "
                + f"-> {rxn_coords[-1]:.2f}"
            )
            self._print(
                f"maxfrc: {maxfrc} Ea_f: {energies[imax]-energies[0]:<8.4f} "
                + f"dE: {energies[-1]-energies[0]:<8.4f}"
            )

        return traj_frames

    def as_dict(self) -> dict:
        """"""
        params = {}

        # self._print(f"{self.setting.backend = }")

        for k, v in dataclasses.asdict(self.setting).items():
            if not k.startswith("_"):
                params[k] = v

        return params


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*

import os
import pathlib
from typing import NoReturn

from . import AbstractPotentialManager


class Cp2kManager(AbstractPotentialManager):

    name = "cp2k"

    implemented_backends = ["cp2k", "cp2k_shell"]
    valid_combinations = (
        # calculator, dynamics
        ("cp2k", "ase"),
        ("cp2k", "cp2k"), 
        ("cp2k_shell", "ase")
    )

    def __init__(self, *args, **kwargs):
        """"""

        return

    def register_calculator(self, calc_params):
        """"""
        # - check backends
        super().register_calculator(calc_params)

        # - some extra params
        command = calc_params.pop("command", None)
        directory = calc_params.pop("directory", pathlib.Path.cwd())

        # NOTE: whether check pp and vdw existence
        #       since sometimes we'd like a dummy calculator
        #       -- convert paths to absolute ones
        conflict_keys =[]
        template = calc_params.pop("template", None)
        if template is not None:
            template = str(pathlib.Path(template).absolute())
            self.calc_params.update(template=template)
            # -- check conflicts
            conflict_keys =["cutoff", "max_scf", "xc"]
            from ase.calculators.cp2k import parse_input
            with open(template, "r") as fopen:
                inp = "".join(fopen.readlines())
            root_section = parse_input(inp)
            # TODO: check if run_type, default is ENERGY_FORCE
            # TODO: cutoff exists?
            _ = root_section.get_subsection("FORCE_EVAL/DFT/MGRID")
            print("FORCE_EVAL/DFT/MGRID: ", _)
            # TODO: max_scf exists?
            # TODO: xc exists?
        else:
            inp = ""

        # - check basis_set and pseudo_potential
        path_keywords = ["basis_set_file", "potential_file"]
        for k in path_keywords:
            fpath = calc_params.pop(k, None)
            if fpath is not None:
                calc_params[k] = str(pathlib.Path(fpath).resolve())
                self.calc_params[k] = str(pathlib.Path(fpath).resolve())

        if self.calc_backend == "cp2k":
            from gdpx.computation.cp2k import Cp2kFileIO as CP2K
        elif self.calc_backend == "cp2k_shell":
            from ase.calculators.cp2k import CP2K as CP2K
        else:
            raise NotImplementedError(f"Unimplemented backend {self.calc_backend} for vasp.")

        # - set some parameters
        calc = CP2K(
            directory=directory, command=command,
        )
        calc.set(inp=inp)
        if "cutoff" in conflict_keys:
            calc.set(cutoff=None) # set in inp
        if "max_scf" in conflict_keys:
            calc.set(max_scf=None)# set in inp
        if "xc" in conflict_keys:
            calc.set(xc=None) # set in inp
        calc.set(**calc_params)
        
        self.calc = calc

        return

if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import itertools

import numpy as np

from ase.io import read, write

from .describer import AbstractDescriber


class SpcDescriber(AbstractDescriber):

    name: str = "spc"

    """Compute some properties.
    """

    def __init__(self, *args, **kwargs) -> None:
        """"""
        super().__init__(*args, **kwargs)

        return

    def run(self, structures, worker, *args, **kwargs):
        """"""
        self._print(f"{structures = }")
        self._print(f"{worker = }")

        status = "unfinished"

        # -
        cache_structures = self.directory / "cache.xyz"
        if not cache_structures.exists():
            worker.run(
                structures,
            )
            worker.inspect(resubmit=True)
            if worker.get_number_of_running_jobs() == 0:
                results = worker.retrieve(include_retrieved=True)
                computed_structures = list(itertools.chain(*results))
                write(cache_structures, computed_structures)
            else:
                self._print("wait worker to finish...")
                computed_structures = None
        else:
            computed_structures = read(cache_structures, ":")

        # -
        if computed_structures is not None:
            prev_energies = np.array(
                [a.get_potential_energy(apply_constraint=False) for a in structures]
            )
            curr_energies = np.array(
                [
                    a.get_potential_energy(apply_constraint=False)
                    for a in computed_structures
                ]
            )
            abs_ene_err = np.fabs(curr_energies - prev_energies)
            #self._print(f"{abs_ene_err = }")

            prev_forces = np.array(
                [a.get_forces(apply_constraint=False).flatten() for a in structures]
            )
            curr_forces = np.array(
                [a.get_forces(apply_constraint=False).flatten() for a in computed_structures]
            )
            max_frc_err = np.max(np.fabs(curr_forces - prev_forces), axis=1)
            #self._print(f"{max_frc_err = }")

            for i, a in enumerate(structures):
                a.info["abs_ene_err"] = abs_ene_err[i]
                a.info["max_frc_err"] = max_frc_err[i]
            
            status = "finished"
        else:
            ...

        return status


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from typing import List

import numpy as np
import numpy.ma as ma
from scipy.spatial import distance_matrix

from joblib import Parallel, delayed

from ase import Atoms
from ase.io import read, write
from ase.neighborlist import NeighborList, natural_cutoffs

from ..builder.group import create_a_group
from .comparator import AbstractComparator
from ..utils.command import CustomTimer


def switch_function(distances, r_cut, r_shift=0., nn: int=6, mm: int=None):
    """"""
    nn = (6 if nn is None else nn)
    mm = (nn*2 if mm is None else mm)

    scaled_distances = (distances - r_shift) / r_cut

    return (1 - scaled_distances**nn) / (1 - scaled_distances**mm)


def compute_self_coorditaion_number(positions):
    """"""
    dmat = distance_matrix(positions, positions)
    #print(dmat)

    sf = switch_function(dmat, r_cut=3.8, nn=8, mm=14)
    np.fill_diagonal(sf, 0.)
    coordination = np.sum(sf, axis=1)

    return coordination

def compute_coorditaion_number(positions, indices_a, indices_b, r_cut, nn=None, mm=None):
    """NOTE: This does not substract the self-coordination."""
    dmat = distance_matrix(positions[indices_a, :], positions[indices_b, :])
    #print(dmat)

    sf = switch_function(dmat, r_cut=r_cut, nn=8, mm=14)
    #print(sf)
    coordination = np.sum(sf, axis=1)
    #print(coordination)

    return coordination

def compute_coordition_number_by_neighbour(nl, cell, positions, indices_a, indices_b, r_cut, nn=None, mm=None):
    """"""
    # NOTE: For the two same groups, the distance matrix may not be square
    #       if self-inetraction is considered since one atom can interact with
    #       several itselves under periodic boundary condition.
    dmat, masks = [], []
    for i in indices_a:
        distances = []
        nei_indices, nei_offsets = nl.get_neighbors(i)
        for j, offset in zip(nei_indices, nei_offsets):
            if j in indices_b:
                distance = np.linalg.norm(
                    positions[i, :] - positions[j, :] + np.dot(offset, cell)
                )
                #print(f"ditsance {i} <-> {j}: {distance}")
                # check if distance is nonzero (as with itself) and under r_cut.
                if 1e-8 < distance <= r_cut:
                    distances.append(distance)
            else:
                ...
        dmat.append(distances)
    coordination = []
    for distances in dmat:
        #print(distances)
        curr_coordination = np.sum(switch_function(np.array(distances), r_cut=r_cut, nn=nn, mm=mm))
        coordination.append(curr_coordination)
    coordination = np.array(coordination)
    #print(coordination)

    return coordination


class CoordinationComparator(AbstractComparator):

    def __init__(self, pairs=None, acn_rmse: float=0.02, acn_max: float=0.20, *args, **kwargs):
        """"""
        super().__init__(*args, **kwargs)

        # - check pairs that should be a List of 2-element Lists
        pairs = (pairs if pairs is not None else [])
        npairs = len(pairs)
        assert npairs > 0, "Need at least 1 pair."
        pairs_ = []
        for i, pair in enumerate(pairs):
            ngroups = len(pair)
            assert ngroups >= 2, f"Need at least two groups to define the pair {i}."
            pairs_.append(pair+[None]*(5-ngroups))
        self.pairs = pairs_

        self.acn_rmse = acn_rmse
        self.acn_max = acn_max

        return
    
    @staticmethod
    def _process_single_structure(atoms, pair_info, *args, **kwargs):
        """"""
        natoms = len(atoms)
        max_r_cut = np.max([pair[2] for pair in pair_info])
        nl = NeighborList(
            cutoffs=[max_r_cut/2.]*natoms, skin=0.2, sorted=False,
            self_interaction=True, bothways=True
        )
        nl.update(atoms)
        coordination = []
        for group_a, group_b, r_cut, nn, mm in pair_info:
            indices_a = create_a_group(atoms, group_a)
            indices_b = create_a_group(atoms, group_b)
            #curr_coordination = compute_coorditaion_number(
            #    atoms.positions, indices_a, indices_b, r_cut, nn, mm
            #)
            curr_coordination = compute_coordition_number_by_neighbour(
                nl, atoms.cell, atoms.positions, indices_a, indices_b, r_cut, nn, mm
            )
            coordination.extend(sorted(curr_coordination.tolist()))

        return coordination
    
    def prepare_data(self, frames: List[Atoms]):
        """"""
        #coordinations = []
        #for i, atoms in enumerate(frames):
        #    self._debug(f"frame {i}")
        #    coordination = self._process_single_structure(atoms, self.pairs)
        #    coordinations.append(coordination)

        with CustomTimer(name="coordination", func=self._debug):
            coordinations = Parallel(n_jobs=self.njobs)(
                delayed(self._process_single_structure)(
                    atoms, self.pairs
                ) for atoms in frames
            )

        coordinations = np.array(coordinations)

        return coordinations
    
    def looks_like(self, fp1, fp2):
        """Compare the fingerprints, here, the coordination number."""
        fp1, fp2 = np.array(fp1), np.array(fp2)

        assert fp1.shape == fp2.shape, f"{fp1.shape} vs. {fp2.shape} ..."
        nfeatures = fp1.shape[0]

        is_similar = False
        
        error = fp1 - fp2
        rmse = np.sqrt(np.sum(error**2)/nfeatures)
        if rmse <= self.acn_rmse and np.max(np.fabs(error)) <= self.acn_max:
            is_similar = True

        return is_similar


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ..core.register import registers
from ..core.variable import Variable
from ..core.operation import Operation
from ..core.variable import DummyVariable


@registers.variable.register
class ComparatorVariable(Variable):

    def __init__(self, directory="./", *args, **kwargs):
        """"""
        method = kwargs.pop("method", None)
        comparator = registers.create("comparator", method, convert_name=False, **kwargs)
        super().__init__(initial_value=comparator, directory=directory)

        return


@registers.operation.register
class compare(Operation):

    status = "finished" # Always finished since it is not time-consuming

    def __init__(self, reference, prediction = DummyVariable(), comparator = DummyVariable(), directory="./") -> None:
        """"""
        super().__init__(input_nodes=[reference, prediction, comparator], directory=directory)

        return
    
    def forward(self, reference, prediction, comparator):
        """"""
        super().forward()

        comparator.directory = self.directory
        comparator.run(prediction, reference)

        return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*


import abc
import copy
import os
import pathlib
import subprocess

from typing import Union, Callable, List

from ..core.node import AbstractNode


class TrainingFailed(RuntimeError):
    """Training unexpectedly fails."""

    ...


class FreezingFailed(RuntimeError):
    """Freezing unexpectedly fails."""

    ...


class AbstractTrainer(AbstractNode):

    #: Name of this trainer.
    name: str = "trainer"

    #: The path of the command executable.
    command: str = None

    #: Command to freeze/deploy.
    freeze_command: str = None

    #: Type list e.g. [C, H, O].
    _type_list: List[str] = None

    #: Prefix of input file.
    prefix: str = "config"

    def __init__(
        self,
        config: dict,
        type_list: List[str] = None,
        train_epochs: int = 200,
        print_epochs: int = 5,
        directory=".",
        command="train",
        freeze_command="freeze",
        random_seed: Union[int, dict] = None,
        *args,
        **kwargs,
    ) -> None:
        """"""
        super().__init__(directory=directory, random_seed=random_seed)
        self.command = command
        if freeze_command is None:
            self.freeze_command = self.command
        else:
            self.freeze_command = freeze_command

        self.directory = directory
        self.config = config  # train model parameters

        # - TODO: sync type_list

        self.train_epochs = train_epochs
        self.print_epochs = print_epochs

        return

    @property
    def directory(self):
        """Directory should always be absolute."""

        return self._directory

    @directory.setter
    def directory(self, directory: Union[str, pathlib.Path]):
        """"""
        self._directory = pathlib.Path(directory).resolve()

        return

    @property
    def type_list(self):
        """"""

        return self._type_list

    @abc.abstractmethod
    def _resolve_train_command(self, *args, **kwargs):
        """"""

        return

    @abc.abstractmethod
    def _resolve_freeze_command(self, *args, **kwargs):
        """"""

        return

    @property
    @abc.abstractmethod
    def frozen_name(self):
        """"""
        ...

    def _train_from_the_scratch(self, dataset, init_model):
        """Train from the scratch."""
        command = self._resolve_train_command(init_model)
        if command is None:
            raise TrainingFailed(
                f"Please supply the command keyword for {self.name.upper()}."
            )

        if not self.directory.exists():
            self.directory.mkdir(parents=True, exist_ok=True)
        self.write_input(dataset)

        return command

    def train(self, dataset, init_model=None, *args, **kwargs):
        """"""
        if not hasattr(self, "_train_from_the_restart"):
            command = self._train_from_the_scratch(dataset, init_model)
        else:
            command = self._train_from_the_restart(dataset, init_model)
        self._print(f"TRAINING COMMAND: {command}")

        try:
            proc = subprocess.Popen(command, shell=True, cwd=self.directory)
        except OSError as err:
            msg = "Failed to execute `{}`".format(command)
            raise TrainingFailed(msg) from err

        errorcode = proc.wait()

        if errorcode:
            path = os.path.abspath(self.directory)
            msg = (
                'Trainer "{}" failed with command "{}" failed in '
                "{} with error code {}".format(self.name, command, path, errorcode)
            )
            raise TrainingFailed(msg)

        return

    def freeze(self):
        """Freeze trained model and return the model path."""
        frozen_model = (self.directory / self.frozen_name).resolve()
        if not frozen_model.exists():
            command = self._resolve_freeze_command()
            try:
                proc = subprocess.Popen(command, shell=True, cwd=self.directory)
            except OSError as err:
                msg = "Failed to execute `{}`".format(command)
                raise FreezingFailed(msg) from err

            errorcode = proc.wait()

            if errorcode:
                path = os.path.abspath(self.directory)
                msg = (
                    'Trainer "{}" failed with command "{}" failed in '
                    "{} with error code {}".format(self.name, command, path, errorcode)
                )
                raise FreezingFailed(msg)
        else:
            ...

        return self.directory / self.frozen_name

    @abc.abstractmethod
    def write_input(self, dataset, *args, **kwargs):
        """Convert dataset to the target format and write the configuration file if it has."""

        return

    @abc.abstractmethod
    def read_convergence(self) -> bool:
        """"""

        return

    def as_dict(self) -> dict:
        """"""
        trainer_params = {}
        trainer_params["name"] = self.name
        trainer_params["type_list"] = self.type_list
        trainer_params["config"] = self.config
        trainer_params["command"] = self.command
        trainer_params["freeze_command"] = self.freeze_command
        trainer_params["train_epochs"] = self.train_epochs

        # NOTE: self.random_seed may be changed thus we save the init one
        trainer_params["random_seed"] = self.init_random_seed

        trainer_params = copy.deepcopy(trainer_params)

        return trainer_params


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import List

import numpy as np

from ase import Atoms
from ase.constraints import constrained_indices


"""Some surface-detection methods."""


def normalize(vector):
    """"""
    return vector / np.linalg.norm(vector) if np.linalg.norm(vector) != 0 else vector * 0


def generate_normals(
    atoms: Atoms, nl,
    surf_indices: List[int] = None,
    ads_indices: List[int] = [],
    mask_elements: List[str] = [],
    surf_norm_min: float = 0.5,
    normalised: bool=True, system: str = "surface"
):
    """Find normals to surface of a structure.

    Atoms that have positive z-axis normal would be on the surface.

    Args:
        atoms: Atoms object.
        nl: Pre-created neighbour list.
        surf_indices: Indices of surface atoms.
        ads_indices: Indices of adsorbate atoms.
        mask_elements: Indices of atoms that are not considered for adsorption sites.
        surf_norm_min: Minimum for a surface normal.
        normalised: Whether normalise normals.

    """
    natoms = len(atoms)

    normals = np.zeros(shape=(natoms, 3), dtype=float)
    for i, atom in enumerate(atoms):
        # print("centre: ", index, atom.position)
        if i in ads_indices:
            continue
        normal = np.array([0, 0, 0], dtype=float)
        for nei_idx, offset in zip(*nl.get_neighbors(i)):
            if nei_idx in ads_indices:
                continue
            # print("neigh: ", neighbor)
            direction = atom.position - (
                atoms[nei_idx].position + np.dot(offset, atoms.get_cell())
            )
            normal += direction
        # print("direction:", normal)
        if np.linalg.norm(normal) > surf_norm_min:
            normals[i, :] = normalize(normal) if normalised else normal

    # NOTE: check normal is pointed to z-axis surface
    if surf_indices is not None:
        surf_indices_ = surf_indices
        for i in range(natoms):
            if i not in surf_indices_:
                normals[i] = np.zeros(3)
    else:
        if system.startswith("surface"):
            surf_indices_ = [
                i
                for i in range(natoms)
                if np.linalg.norm(normals[i]) > 1e-5 and normals[i][2] > 0.0
            ]
        elif system.startswith("cluster"):
            surf_indices_ = [
                i for i in range(natoms) if np.linalg.norm(normals[i]) > 1e-5
            ]
        else:
            raise RuntimeError(f"Surface detection is not available for system {system}.")

    # - remove constrained atoms from surface atoms
    constrained = constrained_indices(atoms)
    surf_indices_ = [i for i in surf_indices_ if i not in constrained]

    # - remove unallowed elements
    surf_indices_ = [i for i in surf_indices_ if atoms[i].symbol not in mask_elements]

    return normals, surf_indices_


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import itertools

from typing import List

import numpy as np

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
try:
    plt.style.use("presentation")
except Exception as e:
    #print("Used default matplotlib style.")
    ...

from ase import Atoms
from ase.io import read, write
from ase.calculators.singlepoint import SinglePointCalculator
from ase.utils.forcecurve import fit_raw

from .validator import AbstractValidator

"""Validate reaction pathway using NEB.
"""

def make_clean_atoms(atoms_):
    """Create a clean atoms from the input."""
    atoms = Atoms(
        symbols=atoms_.get_chemical_symbols(),
        positions=atoms_.get_positions().copy(),
        cell=atoms_.get_cell().copy(),
        pbc=copy.deepcopy(atoms_.get_pbc())
    )

    return atoms

def get_forcefit(images):
    """"""
    energies = np.array([a.get_potential_energy() for a in images])
    emin = np.min(energies)
    forces = [a.get_forces(apply_constraint=True) for a in images]
    positions = [a.positions for a in images]
    cell, pbc = images[0].get_cell(complete=True), True
    #energies -= emin
    #rxn_coords = compute_rxn_coords(images)
    #ax.scatter(rxn_coords, energies, label="dp")
    ff = fit_raw(energies, forces, positions, cell, pbc) # ForceFit
    #print(ff)

    return ff

def plot_results(images_dict: dict, prefix, wdir):
    """"""
    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(12, 8))
    plt.suptitle("Nudged Elastic Band Calculation")

    ax.set_xlabel("Reaction Coordinate [Å]")
    ax.set_ylabel("Potential Energy [eV]")

    #print(mcolors.TABLEAU_COLORS)
    for (name, images), c in zip(images_dict.items(), itertools.cycle(mcolors.TABLEAU_COLORS.keys())):
        ff = get_forcefit(images)

        ax.scatter(ff.path, ff.energies, color=c, facecolor="w", zorder=100)
        ax.plot(ff.fit_path, ff.fit_energies, "-", color=c, label=name)

    np.savetxt(wdir/f"{prefix}neb.dat", np.vstack((ff.fit_path, ff.fit_energies)).T, fmt="%8.4f")

    ax.legend()

    plt.savefig(wdir/f"{prefix}neb.png")

    return

class PathwayValidator(AbstractValidator):


    def run(self, dataset, worker = None, *args, **kwargs):
        """"""
        super().run()

        # - prediction
        #nebtraj = dataset.get("prediction", None)[0]
        nebtraj = dataset.get("prediction", None)
        self._irun(images=nebtraj, prefix="pre-", worker=worker)

        return
    
    def _irun(self, images: List[Atoms], prefix: str, worker):
        """"""
        self._print(f"{worker = }")
        # TODO: Different NEB pathway
        #       default number of images
        #       use its own number
        #       same number as reference
        worker.directory = self.directory/"_xxx"
        worker.run([images[0], images[-1]])
        worker.inspect(resubmit=True)
        if worker.get_number_of_running_jobs() == 0:
            ret = worker.retrieve(
                include_retrieved=True,
            ) # (npairs, nbands, nimages)
            #self._print(f"{ret = }")

        images_dict = dict(
            reference = images,
            prediction = ret[0][-1]
        )

        plot_results(images_dict, prefix=prefix, wdir=self.directory)

        return

    def _compare_results(self, images_dict: dict):
        """"""        

        raise NotImplementedError()


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy

from .. import config
from ..register import registers
from ..operation import Operation

def traverse_postorder(operation):

    nodes_postorder = []
    identifiers = []

    def recurse(node):
        if isinstance(node, Operation):
            for input_node in node.input_nodes:
                recurse(input_node)
        if id(node) not in identifiers:
            nodes_postorder.append(node)
            identifiers.append(id(node))

    recurse(operation)

    return nodes_postorder

def create_variable(node_name, node_params_: dict):
    """"""
    node_params = copy.deepcopy(node_params_)
    node_type = node_params.pop("type", None)
    assert node_type is not None, f"{node_name} has no type."
    node_template = node_params.pop("template", None)
    # -- special keywords
    #random_seed = node_params.pop("random_seed", None)
    #if random_seed is not None:
    #    rng = np.random.default_rng(random_seed)
    #    node_params.update(rng=rng)
    config._debug(node_name)
    config._debug(node_params)

    # -- 
    node = None
    if node_template is not None:
        node_params.update(**node_template)
    node_cls = registers.get("variable", node_type, convert_name=True)
    node = node_cls(**node_params)

    return node

def create_operation(op_name, op_params_: dict):
    """"""
    op_params = copy.deepcopy(op_params_)
    op_type = op_params.pop("type", None)
    assert op_type is not None, f"{op_name} has no type."
    op_template = op_params.pop("template", None)
    # -- special keywords
    #random_seed = op_params.pop("random_seed", None)
    #if random_seed is not None:
    #    rng = np.random.default_rng(random_seed)
    #    op_params.update(rng=rng)
    config._debug(op_name)
    config._debug(op_params)

    # --
    op_cls = registers.get("operation", op_type, convert_name=False)
    operation = op_cls(**op_params)

    return operation


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import abc
import copy
import pathlib
from typing import Optional, List, Mapping

import numpy as np
from sklearn.decomposition import PCA

from ase import Atoms
from ase.io import read, write

from gdpx.core.node import AbstractNode

try:
    from dscribe.descriptors import SOAP
except Exception as e:
    print(e)

USE_CHEMISCOPE = 0
try:
    import chemiscope
    USE_CHEMISCOPE = 1
except Exception as e:
    print(e)


class AbstractDescriber(AbstractNode):

    cache_features = "features.npy"

    def __init__(self, *args, **kwargs) -> None:
        """"""
        super().__init__(*args, **kwargs)

        return

    @abc.abstractmethod
    def run(self, dataset, *args, **kwargs):
        """"""

        return

    def _write_chemiscope(self, frames, features):
        """"""
        # - write chemiscope inputs
        pca = PCA(n_components=2).fit_transform(features)
        properties = dict(
            PCA = dict(
                target = "structure",
                values = pca
            ),
            #energies = dict(
            #    target = "structure",
            #    values = [a.get_potential_energy() for a in frames],
            #    units = "eV"
            #)
        )

        frame_properties = chemiscope.extract_properties(
            frames,
            only=["energy"]
        )
        properties.update(**frame_properties)

        chemiscope.write_input(
            self.directory/"my-input.json.gz", frames=frames, properties=properties
        )

        return

    def _plot_results(self, features, groups: Mapping[str,List[int]], *args, **kwargs):
        """"""
        # - plot selection
        import matplotlib.pyplot as plt
        try:
            plt.style.use("presentation")
        except Exception as e:
            ...

        reducer = PCA(n_components=2)
        reducer.fit(features)

        for i, (name, indices) in enumerate(groups.items()):
            proj = reducer.transform(features[indices,:])
            plt.scatter(
                proj[:, 0], proj[:, 1], alpha=0.5, zorder=100-i,
                label=f"{name} {len(indices)}"
            )

        plt.legend()
        plt.axis("off")
        plt.savefig(self.directory/"pca.png")

        return 


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from typing import List, Callable

import numpy as np
import jax.numpy as jnp

from ase.calculators.morse import MorsePotential
from ase.neighborlist import neighbor_list

from .bias import AbstractBias



def morse(positioins, cvfunc: Callable):
    """"""

    return


class NucleiBias(AbstractBias):

    implemented_properties = ["energy", "free_energy", "forces"]

    default_parameters = dict()

    def __init__(self, colvar: dict=None, restart=None, label=None, atoms=None, directory=".", **kwargs):
        """"""
        super().__init__(colvar=colvar, restart=restart, label=label, atoms=atoms, directory=directory, **kwargs)

        return

    def calculate(self, atoms=None, properties=["energy"], system_changes=["positions"]):
        """"""
        super().calculate(atoms, properties, system_changes)

        positions = jnp.array(atoms.get_positions())
        ret = self._compute_bias(
            positions, self.colvar,
            **self._compute_params
        )
        self.results["energy"] = np.asarray(ret[0])
        self.results["forces"] = -np.array(ret[1]) # copy forces

        return


if __name__ == "__main__":
    ...


#!/usr3/bin/env python3
# -*- coding: utf-8 -*


import numpy as np

from ase import units

try:
    import plumed
except Exception as e:
    print(e)


class AsePlumed(object):

    def __init__(self, atoms, timestep, inp_file="plumed.dat", out_file="plumed.out"):

        self.atoms = atoms
        self.timestep = timestep
        self.natoms = len(atoms)
        self.masses = self.atoms.get_masses().copy()  # masses cannot change

        self.inp_file = inp_file
        self.out_file = out_file

        self.worker = self.initialize()

        return

    def initialize(self):
        # init
        p_md = plumed.Plumed()

        # units
        energyUnits = units.mol / units.kJ  # eV to kJ/mol
        lengthUnits = 1.0 / units.nm  # angstrom to nm
        timeUnits = 1.0 / (1000.0 * units.fs)  # fs to ps

        p_md.cmd("setMDEnergyUnits", energyUnits)
        p_md.cmd("setMDLengthUnits", lengthUnits)
        p_md.cmd("setMDTimeUnits", timeUnits)

        # inp, out
        p_md.cmd("setPlumedDat", self.inp_file)
        p_md.cmd("setLogFile", self.out_file)

        # simulation details
        p_md.cmd("setTimestep", self.timestep)
        p_md.cmd("setNatoms", self.natoms)
        p_md.cmd("setMDEngine", "ase")

        # finally!
        p_md.cmd("init")

        return p_md

    def external_forces(
        self,
        istep,
        new_energy=None,
        new_forces=None,  # sometimes use forces not attached to self.atoms
        new_virial=None,
        delta_forces=False,
    ):
        """return external forces from plumed"""
        # structure info
        positions = self.atoms.get_positions().copy()
        cell = self.atoms.cell[:].copy()

        if new_forces is None:
            forces = self.atoms.get_forces().copy()
        else:
            forces = new_forces.copy()
        original_forces = forces.copy()

        if new_energy is None:
            energy = self.atoms.get_potential_energy()
        else:
            energy = new_energy

        # TODO: get virial
        virial = np.zeros((3, 3))

        self.worker.cmd("setStep", istep)
        self.worker.cmd("setMasses", self.masses)
        self.worker.cmd("setForces", forces)
        self.worker.cmd("setPositions", positions)
        self.worker.cmd("setEnergy", energy)
        self.worker.cmd("setBox", cell)
        self.worker.cmd("setVirial", virial)
        self.worker.cmd("calc", None)

        # implent plumed external forces into momenta
        if delta_forces:
            plumed_forces = forces - original_forces
        else:
            plumed_forces = forces

        return plumed_forces

    def finalize(self):
        self.worker.finalize()


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from itertools import groupby
import time

import numpy as np
import scipy as sp

from gdpx.selector.abstract import create_selector
from gdpx.utils.comparasion import parity_plot_dict, rms_dict
from gdpx.data.operators import append_predictions, merge_predicted_forces, xyz2results
from ase.calculators.singlepoint import SinglePointCalculator
from ase.constraints import FixAtoms
from ase.io import read, write
from ase import Atoms
from tqdm import tqdm
import pathlib
from pathlib import Path

from typing import NoReturn, Union, List, overload
from joblib import Parallel, delayed

from collections import Counter

from gdpx.utils.command import parse_input_file

# global settings
from gdpx import config


import matplotlib as mpl
mpl.use("Agg") #silent mode
from matplotlib import pyplot as plt
try:
    plt.style.use("presentation")
except Exception as e:
    #print("Used default matplotlib style.")
    ...


"""
properties we may be interested in

maximum force if it is a local minimum

energy and force distribution (both train and test)

use DeepEval to check uncertainty

"""

def estimate_number_composition():
    mini = 320
    base = 32


    for coef in range(1,19):
        #print(mini*coef**0.5)
        print(np.ceil(mini*coef**0.5/base)*base)

    return


def transform_forces(symbols, forces):
    """"""
    type_list = ["O", "Pt"]
    type_map = {"O": 0, "Pt": 1}
    # print(symbols)
    forces = np.array(forces)
    print(forces.shape)
    elemental_forces = {}
    for elem in type_map.keys():
        elemental_forces[elem] = []
    for idx, symbol in enumerate(symbols):
        # print(type_list[symbol])
        elemental_forces[type_list[symbol]].extend(
            list(forces[:, 3*idx:3*(idx+1)].flatten())
        )
    for elem in elemental_forces.keys():
        elemental_forces[elem] = np.array(elemental_forces[elem])
        print(elemental_forces[elem].shape)
    elemental_forces = elemental_forces

    return elemental_forces


def plot_hist(ax, data, xlabel, ylabel, label=None):
    num_bins = 50
    n, bins, patches = ax.hist(data, num_bins, density=False, label=label)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    return

def parse_constraint_info(cons_info):
    """ python convention starts from 0 and excludes end
        lammps convention starts from 1 and includes end
    """
    cons_indices = None
    if cons_info is not None:
        print("apply constraint...")
        cons_indices = []
        for c in cons_info.split():
            s, e = c.split(":")
            cons_indices.extend(list(range(int(s)-1, int(e))))

    return

class DatasetSystems():

    """ PlainTextDatasetSystem
    """

    def __init__(
        self,
        system_config_file
    ):
        input_dict = parse_input_file(system_config_file)
        self.database_path = Path(input_dict["database"])
        self.systems = input_dict["systems"]
        self.sift_settings = input_dict["sift"]

        for sys_name in self.systems.keys():
            self.__check_system_name(sys_name, self.systems[sys_name]["composition"])

        return
    
    def __check_system_name(self, sys_name, composition):
        """ system name should be one of three formats
            1. simple chemical formula
            2. custom name plus chemical formula
            3. custom name, chemical formula and system type
        """
        name_parts = sys_name.split("-")
        if len(name_parts) == 1:
            chemical_formula = name_parts[0]
        elif len(name_parts) == 2:
            custom_name, chemical_formula = name_parts
        elif len(name_parts) == 3:
            custom_name, chemical_formula, system_type = name_parts
        else:
            raise ValueError("directory name must be as xxx, xxx-xxx, or xxx-xxx-xxx")

        composition_from_name = self.__parse_composition_from_name(chemical_formula)
        assert composition_from_name == composition, f"{sys_name}'s composition is not consistent."

        return

    def __parse_composition_from_name(self, chemical_formula: str):
        """"""
        from itertools import groupby
        formula_list = [
            "".join(g) for _, g in groupby(chemical_formula, str.isalpha)
        ]
        iformula = iter(formula_list)
        number_dict = dict(zip(iformula, iformula))  # number of atoms per type

        type_list = []
        for atype, num in number_dict.items():
            number_dict[atype] = int(num)
            if number_dict[atype] > 0:
                type_list.append(atype)
        assert len(type_list) > 0, "no atomic types were found."

        return number_dict

    def read_all_systems(
        self, system_path,
        sys_pattern = "*", file_pattern = "*.xyz"
    ) -> List[Atoms]:
        print("===== load all systems =====")
        # TODO: should sort dirs to make frames consistent
        #system_path = self.database_path
        system_path = Path(system_path)
        print(system_path)

        nframes = 0
        total_frames = {}
        #dirs = list(system_path.glob(sys_pattern)) 
        dirs = []
        for sys_name in self.systems.keys():
            dirs.append(system_path / sys_name)
            print(system_path / sys_name)
        dirs.sort()
        for p in dirs:
            cur_frames = self.read_single_system(p, file_pattern)
            total_frames[p.name] = cur_frames
            nframes += len(cur_frames)
        print("===== Total number: ", nframes)

        return total_frames

    def read_single_system(
        self, system_path: Union[str, pathlib.Path],
        pattern: int = "*.xyz"
    ) -> List[Atoms]:
        print("----- read frames -----")
        print("system path: ", str(system_path))
        # TODO: should sort dirs to make frames consistent

        # TODO: check composition is the same
        total_frames = []
        stru_files = list(system_path.glob(pattern)) # structure files, xyz for now
        print("wocao: ", stru_files)
        stru_files.sort()
        for p in stru_files:
            frames = read(p, ":")
            total_frames.extend(frames)
            print("structure path: ", p)
            print("number of frames: ", len(frames))
        print("Total number: ", len(total_frames))

        # sift structures based on atomic energy and max forces
        sift_criteria = self.sift_settings
        use_sift = sift_criteria.get("enabled", False)
        
        if use_sift:
            total_frames = self.sift_structures(
                total_frames,
                energy_tolerance = sift_criteria["atomic_energy"],
                force_tolerance = sift_criteria["max_force"]
            )
        exit()

        return total_frames

    def sift_structures(
        self, 
        frames,
        energy_tolerance = None, 
        force_tolerance = None
    ) -> NoReturn:
        """ sift structures based on atomic energy and max forces
        """
        # remove large-force frames
        print("----- sift structures -----")
        if force_tolerance is None and energy_tolerance is None:
            print("no criteria is applied...")
        else:
            if energy_tolerance is None:
                energy_tolerance = 1e8
            if force_tolerance is None:
                force_tolerance = 1e8
            print("energy criteria: ", energy_tolerance, " [eV]")
            print("force criteria: ", force_tolerance, " [eV/Å]")
            new_frames = []
            for idx, atoms in enumerate(frames):
                avg_energy = atoms.get_potential_energy() / len(atoms)
                max_force = np.max(np.fabs(atoms.get_forces()))
                if max_force > force_tolerance: 
                    print(f"skip {idx} with large forces {max_force}")
                    continue
                elif avg_energy > energy_tolerance:
                    print(f"skip {idx} with large energy {avg_energy}")
                    continue
                else:
                    new_frames.append(atoms)
            print("number of frames after sift: ", len(new_frames))

        return new_frames
    
    def __str__(self):
        content = "===== DataSystems =====\n"
        content += f"nsystems {len(self.systems.keys())}\n"
        content += "===== End DataSystems =====\n"

        return content

class DataOperator():

    """
    """

    def __init__(
        self, main_dir, 
        systems, global_type_list,
        name: str,
        pattern, 
        calc,
        geometric_settings,
        sift_settings,
        compress_settings = None,
        selection_settings = None
    ):
        """ parse names...
        """
        # get global type list
        self.calc = calc

        # TODO: in a simpler way...
        self.sift_settings = sift_settings

        # geometric convergence
        self.geometric_settings = geometric_settings
        self.fmax = geometric_settings.get("fmax", 0.05)

        # compress settings
        self.compress_params = compress_settings

        self.name = name # system name

        name_parts = name.split("-")
        if len(name_parts) == 1:
            chemical_formula = name_parts[0]
        elif len(name_parts) == 2:
            custom_name, chemical_formula = name_parts
        elif len(name_parts) == 3:
            custom_name, chemical_formula, system_type = name_parts
        else:
            raise ValueError("directory name must be as xxx-xxx-xxx")

        if name == "ALL":
            self.type_list, self.type_numbers = [], []
        else:
            self.type_list, self.type_numbers = self.__parse_type_list(
                chemical_formula)

        # TODO: results dir
        self.prefix = Path.cwd() / self.name

        # selection settings
        if selection_settings is not None:
            self.selector = create_selector(selection_settings, directory=self.prefix)

        # --- system specific ---
        self.systems = systems # contains composition and constraint

        # parse structures
        self.frames = None  # List of Atoms
        self.main_dir = main_dir
        self.pattern = pattern
        if name != "ALL":
            self.is_single = True
            self.system_names = [self.name]
            #self.frames = self.read_single_system(
            #    Path(main_dir) / name,
            #    pattern = pattern
            #)
        else:
            self.is_single = False # whether a single dataset is read
            self.system_names = list(self.systems.keys())
            #self.global_type_list = []
            #for sys_name in self.system_names:
            #    # update type_list
            #    chemical_formula = "".join(
            #        [k+str(v) for k, v in self.systems[sys_name]["composition"].items()]
            #    )
            #    self.type_list, self.type_numbers = self.__parse_type_list(
            #        chemical_formula
            #    )
            #    self.global_type_list.extend(self.type_list)
            #self.global_type_list = sorted(list(set(self.global_type_list)))
        self.global_type_list = global_type_list

        # check system accessibility
        accessible_sys_names = []
        for sys_name in self.systems.keys():
            sys_path = self.main_dir / sys_name
            if sys_path.exists():
                nxyzfiles = len(list(sys_path.glob("*.xyz")))
                if nxyzfiles > 0:
                    accessible_sys_names.append(sys_name)
                else:
                    print(f"xxx Empty {sys_path}. xxx")
            else:
                print(f"xxx Cant find {sys_path}. xxx")
        acc_systems = {key: value for key, value in self.systems.items() if key in accessible_sys_names}
        self.systems = acc_systems

        for name in self.systems.keys():
            print(name)

        return

    def __parse_type_list(self, chemical_formula: str):
        from itertools import groupby
        formula_list = [
            "".join(g) for _, g in groupby(chemical_formula, str.isalpha)
        ]
        iformula = iter(formula_list)
        number_dict = dict(zip(iformula, iformula))  # number of atoms per type

        type_list = []
        for atype, num in number_dict.items():
            number_dict[atype] = int(num)
            if number_dict[atype] > 0:
                type_list.append(atype)
                assert atype in self.global_type_list, f"atype {atype} not in global type list."
        assert len(type_list) > 0, "no atomic types were found."

        return type_list, number_dict

    def read_all_systems(
        self, system_path: Union[str, pathlib.Path],
        sys_pattern = "*",
        file_pattern = "*.xyz"
    ) -> List[Atoms]:
        print("----- read frames -----")
        print(str(system_path))
        # TODO: should sort dirs to make frames consistent

        nframes = 0
        total_frames = {}
        #dirs = list(system_path.glob(sys_pattern)) 
        dirs = []
        for sys_name in self.systems.keys():
            dirs.append(system_path / sys_name)
        dirs.sort()
        for p in dirs:
            cur_frames = self.read_single_system(p, file_pattern)
            total_frames[p.name] = cur_frames
            nframes += len(cur_frames)
        print("===== Total number: ", nframes)

        return total_frames

    def read_single_system(
        self, system_path: Union[str, pathlib.Path],
        pattern: int = "*.xyz"
    ) -> List[Atoms]:
        print("----- read frames -----")
        print("system path: ", str(system_path))
        # TODO: should sort dirs to make frames consistent

        # TODO: check composition is the same
        total_frames = []
        stru_files = list(system_path.glob(pattern)) # structure files, xyz for now
        #print(pattern, " wocao: ", stru_files)
        stru_files.sort()
        for p in stru_files:
            frames = read(p, ":")
            total_frames.extend(frames)
            print("structure path: ", p)
            print("number of frames: ", len(frames))
        print("Total number: ", len(total_frames))

        # TODO: assert frames have same composition and lattice

        # sift structures based on atomic energy and max forces
        sift_criteria = self.sift_settings
        use_sift = sift_criteria.get("enabled", False)
        
        if use_sift:
            total_frames = self.sift_structures(
                total_frames,
                energy_tolerance = sift_criteria["atomic_energy"],
                force_tolerance = sift_criteria["max_force"]
            )

        return total_frames

    def sift_structures(
        self, 
        frames,
        energy_tolerance = None, 
        force_tolerance = None
    ) -> NoReturn:
        """ sift structures based on atomic energy and max forces
        """
        # remove large-force frames
        print("----- sift structures -----")
        if force_tolerance is None and energy_tolerance is None:
            print("no criteria is applied...")
        else:
            if energy_tolerance is None:
                energy_tolerance = 1e8
            if force_tolerance is None:
                force_tolerance = 1e8
            print("energy criteria: ", energy_tolerance, " [eV]")
            print("force criteria: ", force_tolerance, " [eV/Å]")
            new_frames = []
            for idx, atoms in enumerate(frames):
                avg_energy = atoms.get_potential_energy() / len(atoms)
                max_force = np.max(np.fabs(atoms.get_forces()))
                if max_force > force_tolerance: 
                    print(f"skip {idx} with large forces {max_force}")
                    continue
                elif avg_energy > energy_tolerance:
                    print(f"skip {idx} with large energy {avg_energy}")
                    continue
                else:
                    new_frames.append(atoms)
            print("number of frames after sift: ", len(new_frames))

        return new_frames

    def split_frames(self, count=0):
        merged_indices = []
        for i in range(count):
            previous_indices_path = pathlib.Path("r" + str(i) + "-indices.npy")
            indices = np.load(previous_indices_path).tolist()
            merged_indices.extend(indices)
        assert len(merged_indices) == len(set(merged_indices)), "Have duplicated structures... {0} != {1}".format(
            len(merged_indices), len(set(merged_indices))
        )
        used_frames, other_frames = [], []
        for i, atoms in enumerate(self.frames):
            if i in merged_indices:
                used_frames.append(atoms)
            else:
                other_frames.append(atoms)
        print(
            "Number of used frames: ",
            len(used_frames),
            "Number of other frames: ",
            len(other_frames)
        )

        return used_frames, other_frames

    def reduce_dataset(self, number, count, name, energy_shift):
        """"""
        frames = self.frames
        # write(patword+"-tot.xyz", frames)
        prefix = "r" + str(count) + "-"
        if count == 0:
            if number > 0:
                converged_frames = []
                converged_indices = []
                # add converged structures
                for idx, atoms in enumerate(frames):
                    # add constraint to neglect forces of frozen atoms
                    cons = FixAtoms(
                        indices=[a.index for a in atoms if a.position[2] < 2.5])
                    atoms.set_constraint(cons)
                    max_force = np.max(np.fabs(atoms.get_forces()))
                    if max_force < 0.05:
                        converged_frames.append(atoms)
                        converged_indices.append(idx)

                # select frames
                nconverged = len(converged_frames)
                print("Number of converged: ", nconverged)
                features = calc_desc(frames)
                selected_frames = select_structures(
                    name, features, number-nconverged, prefix=prefix,
                    manually_selected=converged_indices
                )  # include manually selected structures
                # selected_frames.extend(converged_frames)
                # TODO: sometims the converged one will be selected... so duplicate...

                if energy_shift > 0:
                    es = energy_shift
                    print(
                        f"add {es} [eV] energy correction to each structure!!!")
                    for atoms in selected_frames:
                        new_energy = atoms.get_potential_energy() + es
                        new_forces = atoms.get_forces(
                            apply_constraint=False).copy()
                        calc = SinglePointCalculator(
                            atoms, energy=new_energy, forces=new_forces)
                        atoms.calc = calc

                print("Writing structure file... ")
                write((prefix+'-sel.xyz'), selected_frames)
                print("")
        else:
            merged_indices = []
            for i in range(count):
                previous_indices_path = pathlib.Path(
                    "r" + str(i) + "-indices.npy")
                indices = np.load(previous_indices_path).tolist()
                merged_indices.extend(indices)
            print(
                "Number of unique frames: ",
                len(set(merged_indices))
            )

            if previous_indices_path.exists() and args.more:
                features_path = cwd / "features.npy"
                features = np.load(features_path)

                assert len(frames) == features.shape[0]

                index_map = []
                rest_features = []
                rest_frames = []
                for idx, atoms in enumerate(frames):
                    if idx not in merged_indices:
                        rest_frames.append(atoms)
                        rest_features.append(features[idx])
                        index_map.append(idx)
                if len(rest_frames) < number:
                    raise ValueError("Not enough structures left...")
                rest_features = np.array(rest_features)
                selected_frames = select_structures(
                    args.name, rest_features, number, index_map=index_map, prefix=prefix)

                print("Writing structure file... ")
                write(cwd / (prefix+patword+'-sel.xyz'), selected_frames)
                print("")

        return

    def find_converged(
        self, sys_name, sys_frames, res_dir
    ):
        # parse constraint indices
        cons_info = self.systems[sys_name].get("constraint", None)
        cons_type, cons_data = None, None
        if cons_info is not None:
            if isinstance(cons_info, dict):
                cons_type = "region"
                cons_data = float(cons_info["zmax"])
            else:
                cons_type = "index"
                if cons_info is not None:
                    print("apply constraint...")
                    cons_indices = []
                    for c in cons_info.split():
                        s, e = c.split(":")
                        cons_indices.extend(list(range(int(s)-1, int(e))))
                cons_data = cons_indices
        print(f"cons: {cons_type} data: {cons_data}")

        # check converged structures
        converged_indices = []
        converged_frames = []
        for i, atoms in enumerate(sys_frames):
            # check if it has constraint
            if cons_type is not None:
                if cons_type == "index":
                    cons = FixAtoms(indices=cons_data)
                    atoms.set_constraint(cons)
                elif cons_type == "region":
                    cons_indices = []
                    for ci, pz in enumerate(atoms.positions[:,2]):
                        if pz < cons_data:
                            cons_indices.append(ci)
                    cons = FixAtoms(indices=cons_indices)
                    atoms.set_constraint(cons)
                else:
                    raise RuntimeError(f"Unknown constraint {cons_type}")
            else:
                # TODO: attached constraint
                cons = atoms.constraints
                if cons:
                    assert len(cons) == 1, "found two constraints on one atoms object"
                    assert isinstance(cons[0], FixAtoms), "the constraint is not FixAtoms"
                else:
                    # no constraints
                    pass
            max_force = np.max(np.fabs(atoms.get_forces(apply_constraint=True)))
            if (max_force < self.fmax):
                atoms.info["description"] = "local minimum"
                converged_frames.append(atoms)
                converged_indices.append(i)
        print("converged structure index: ", converged_indices)

        if len(converged_frames) > 0 and self.geometric_settings["enabled"]:
            converged_energies = [a.get_potential_energy() for a in converged_frames]
            converged_frames.sort(key=lambda a:a.get_potential_energy())
            write(res_dir / "local_minimum.xyz", converged_frames)

        return converged_frames, converged_indices
    
    def compress_systems(
        self,
        minisize = 640, # minimum extra number of structures
        energy_tolerance=0.020,
        energy_shift=0.0
    ):
        # check if selector exists
        if not hasattr(self, "selector"):
            raise RuntimeError("selector is not created...")

        if not self.is_single:
            # save data
            title = ("#{:<23s}  "+"{:<12s}  "*4+"\n").format(
                "SysName", "natoms", "nframes", "ncandidates", "nselected"
            )
            with open(Path.cwd() / "all-compress.dat", "w") as fopen:
                fopen.write(title)

        for sys_name in self.systems.keys():
            # run operation
            print(f"===== Current System {sys_name} =====")
            # update type_list and sys_name
            chemical_formula = "".join(
                [k+str(v) for k, v in self.systems[sys_name]["composition"].items()]
            )
            self.type_list, self.type_numbers = self.__parse_type_list(
                chemical_formula
            )
            natoms = sum(self.type_numbers.values())
            #global_type_list.extend(self.type_list)
            self.name = sys_name

            nframes, ncandidates, nselected, nconverged = self.compress_single_system(
                sys_name, minisize, energy_tolerance, energy_shift
            )

            if not self.is_single:
                # save data
                content = ("{:<24s}  "+"{:<12d}  "*5+"\n").format(
                    sys_name, natoms, nframes, nconverged, ncandidates, nselected
                )
                # TODO: check all
                #all_file = Path.cwd() / "all-compress.dat"
                with open(Path.cwd() / "all-compress.dat", "a") as fopen:
                    fopen.write(content)

        return

    def compress_single_system(
        self,
        sys_name,
        minisize = 640, # minimum extra number of structures
        energy_tolerance=0.020,
        energy_shift=0.0
    ):
        """ select candidates based on 
            true error or model ensemble deviation

            possible criteria 
            energy error 0.020 eV / atom
        """
        nselected, nconverged, ncandidates = 0, 0, 0

        # update a few outputs path based on system
        self.prefix = Path.cwd() / sys_name
        if self.prefix.exists():
            pass
        else:
            self.prefix.mkdir()
        self.selector.directory = self.prefix

        mlp_file = self.prefix / (self.name + "-MLP.xyz")
        out_xyz = self.prefix / ("miaow-sel.xyz")
        if out_xyz.exists():
            print(f"result dir {self.prefix} exists, may skip calculation...")
            return 0, 0, 0

        if self.calc is not None:
            print("!!!use calculator-assisted dataset compression...!!!")
            calc = self.calc

            batchsize = self.compress_params["batchsize"]

            print(f"tolerance {energy_tolerance} shift {energy_shift}")

            if not mlp_file.exists():
                print("\n\n--- calculate mlp results ---")
                # read frames
                frames = self.read_single_system(self.main_dir / sys_name, self.pattern)
                nframes = len(frames)
                
                if nframes > 0:
                    ref_energies, dft_forces = xyz2results(frames, calc=None)

                    #print(frames[0].get_potential_energy())

                    calc_frames = [a.copy() for a in frames] # NOTE: shallow copy of frames will contaminate results
                    #print(frames[0].get_potential_energy())
                    mlp_energies, mlp_forces = xyz2results(calc_frames, calc=calc)

                    # preprocess
                    ref_energies = np.array(ref_energies)
                    mlp_energies = np.array(mlp_energies)
            else:
                frames = read(mlp_file, ":") # TODO: change this into another function
                calc_name = calc.name.lower()

                ref_energies, ref_forces = xyz2results(frames, calc=None)

                mlp_energies = np.array([a.info[calc_name+"_energy"] for a in frames])
                # mlp_forces = merge_predicted_forces(frames, calc_name) # save time for now
            nframes = len(frames)
            if nframes > 0:
                # TODO: check force errors
                natoms_per_structure = np.sum(list(self.type_numbers.values()))

                print("\n\n--- data statistics ---")
                abs_error = np.fabs(ref_energies - mlp_energies) / natoms_per_structure
                print(abs_error)
                print("error shape:")
                print(abs_error.shape)
                print("mean: ", np.mean(abs_error))
                _hist, _bin_edges = np.histogram(abs_error, bins=10)
                print("histogram: ")
                print(_hist)
                print(_bin_edges)

                cand_indices = []  # structure index with large error
                cand_frames = []
                for i, x in enumerate(abs_error):
                    if x > energy_tolerance:
                        cand_indices.append(i)
                        # NOTE: copy atoms wont copy calc
                        _energy = frames[i].get_potential_energy()
                        _forces = frames[i].get_forces()
                        cand_atoms = frames[i].copy()
                        # print(cand_atoms.calc)
                        singlepoint = SinglePointCalculator(
                            cand_atoms, energy=_energy,
                            forces=_forces
                        )
                        cand_atoms.calc = singlepoint
                        # print(cand_atoms.calc)
                        # print(cand_atoms.get_potential_energy())
                        # cand_atoms.calc = None # torch results cannot be direct to joblib
                        cand_frames.append(cand_atoms)
                ncand = len(cand_indices)
                print("number of candidates: ", ncand)

                num_cur = int(
                    np.min([minisize, np.floor(ncand / batchsize)*batchsize]))
                print("number to select: ", num_cur)

                # go for cur
                if num_cur <= 0:
                    print("no selection...")
                    ncandidates, nselected = 0, 0
                else:
                    ncandidates = len(cand_frames)
                    nselected, nconverged = self.compress_system_based_on_structural_diversity(cand_frames, num_cur, energy_shift)
        else:
            print("!!!use descriptor-based dataset compression...!!!")
            # read frames
            sys_frames = self.read_single_system(self.main_dir / sys_name, self.pattern)
            nframes = len(sys_frames)
            if nframes > 0:
                ncandidates = nframes
                nselected, nconverged = self.compress_system_based_on_structural_diversity(sys_frames, minisize, energy_shift)

        # save data
        if nframes > 0:
            title = ("#{:<23s}  "+"{:<12s}  "*4+"\n").format(
                #"SysName", "natoms", "nframes", "dE_RMSE", "dE_Std"
                "SysName", "natoms", "nframes", "ncandidates", "nselected"
            )
            content = ("{:<24s}  "+"{:<12d}  "*4+"\n").format(
                self.name, sum(self.type_numbers.values()), 
                nframes, ncandidates, nselected
            )
            with open(self.prefix / "compress.dat", "w") as fopen:
                fopen.write(title+content)
        else:
            ncandidates, nselected = 0, 0

        return nframes, ncandidates, nselected, nconverged

    def compress_system_based_on_structural_diversity(
        self, 
        frames,
        number = 320,
        energy_shift = 0.0
    ):
        """ Strategy for compressing large dataset
            1. if atomic energy is way large, skip them
            2. random selection based on (
                cur_scores * bin_prop * boltzmann
                # geometric, phase volume, energy-wise
            )
        """

        print("\n\n")
        print("!!! number of frames to compress: ", len(frames))

        #print(frames[0].info)
        #write("fuccc.xyz", frames)
        #print(frames[0].get_potential_energy())

        # find converged structures TODO: move this outside
        converged_frames, converged_indices = self.find_converged(self.name, frames, self.prefix)
        nconverged = len(converged_indices)
        print("Number of converged: ", nconverged)

        # cur in each bin
        print("\n\n--- start histogram ---")
        energies = np.array([a.get_potential_energy() for a in frames])
        # convert to atomic energies
        energies = energies / np.sum(list(self.type_numbers.values()))

        en_width = self.compress_params["en_width"]  # energy width
        en_min = np.floor(np.min(energies))
        en_max = np.ceil(np.max(energies))
        bin_edges = np.arange(en_min, en_max+en_width, en_width)
        nbins = len(bin_edges) - 1  # number of bins

        print(f"AtomicEnergy Range: {en_min} _ {en_max}")
        print(f"number of bins: {nbins}")

        _, bin_edges, binnumber = sp.stats.binned_statistic(
            energies, energies, statistic="mean", bins=bin_edges,
        )

        # add boltzmann coefficient
        beta = self.compress_params["beta"]  # eV
        shifts = self.compress_params["shifts"]
        boltz_prob = np.zeros(len(bin_edges)-1)
        for es in shifts:
            boltz_prob += np.exp(-beta*(bin_edges[:-1]-(en_min+es)))
        norm_boltz = boltz_prob / np.sum(boltz_prob)
        print(f"Boltzmann {boltz_prob}, norm: {norm_boltz}")

        # get frame index in each bin
        fidx_list = []  # frame index list
        for x in range(nbins):
            fidx_list.append([])
        for k, g in groupby(enumerate(binnumber), lambda x: x[1]):
            # k starts with 1 since histogram uses right boundary
            fidx_list[k-1].extend([ig[0] for ig in g])
        num_array = np.zeros(nbins, dtype=int)
        for x in range(nbins):
            num_array[x] = len(fidx_list[x])
        num_density = num_array / np.sum(num_array)
        print(f"NumDensity: {num_array}, norm: {num_density}")

        assert len(frames) == np.sum(
            num_array), "all frames must be in the hist."

        # choose number of structures in eacn bin
        merged_probs = num_density*norm_boltz
        merged_probs = merged_probs / np.sum(merged_probs)
        print("BoltzDensity: ", merged_probs)
        print("random choice: ", number-nconverged)
        if number > nconverged:
            selected_indices = np.random.choice(
                nbins, number-nconverged, replace=True, p=merged_probs)

            final_num_array = np.zeros(nbins, dtype=int)
            for k, g in groupby(selected_indices):
                final_num_array[k] += len(list(g))
            print("Numbers selected in each bin: ", final_num_array)

            # select structures in each bin
            print("\n\n--- start cur decompostion ---")
            # check if selected numbers are valid
            new_final_selected_num = []
            rest_num = 0
            for bin_i, (bin_num, bin_idx) in enumerate(zip(final_num_array, fidx_list)):
                nframes_bin = len(bin_idx) # number of structure in the bin
                #if bin_num == 0:
                #    new_final_selected_num.append(bin_num)
                #    continue
                bin_num += rest_num
                if bin_num > nframes_bin:
                    rest_num = bin_num - nframes_bin
                    bin_num = nframes_bin
                else:
                    rest_num = 0
                #if rest_num < 0:
                #    bin_num -= rest_num
                #rest_num = nframes_bin - bin_num # all minus selected
                #if rest_num < 0:
                #    bin_num = nframes_bin
                new_final_selected_num.append(bin_num)
            print("NEW: Numbers selected in each bin: ", new_final_selected_num)
            all_indices = []
            for bin_i, (bin_num, bin_idx) in enumerate(zip(new_final_selected_num, fidx_list)):
                nframes_bin = len(bin_idx)
                if bin_num == 0 or nframes_bin == 0:
                    continue
                print(
                    "{} candidates are selected from {} structures in bin {}...".format(
                        bin_num, nframes_bin, bin_i
                    )
                )
                #bin_frames = []
                #for i in bin_idx:
                #    bin_frames.append(frames[i])
                bin_frames = [frames[i] for i in bin_idx]
                #features = self.selector.calc_desc(bin_frames)
                #if len(bin_frames) == 1:
                #    features = features[:, np.newaxis]
                #    selected_indices = [0]
                #else:
                #    selected_indices = self.selector.select_structures(
                #        features, bin_num
                #    )
                selected_indices = self.selector.select(bin_frames, ret_indices=True)
                # NOTE: map selected into global indices
                selected_indices = [bin_idx[s] for s in selected_indices]
                    # TODO: sometims the converged one will be selected... so duplicate...
                all_indices.extend(selected_indices)
            all_indices.extend(converged_indices)
        else:
            print("nconverged is greater than number...")
            converged_energies = np.array([a.get_potential_energy() for a in converged_frames])
            converged_probs = -converged_energies + np.max(converged_energies) + 0.1
            converged_probs = converged_probs / np.sum(converged_probs)
            all_indices = np.random.choice(
                converged_indices, number, replace=False, p=converged_probs
            )
        print("number selected: ", len(all_indices))
        print("number no-duplicated: ", len(set(all_indices)))

        selected_frames = []
        for i in all_indices:
            selected_frames.append(frames[i])

        # output final candidates
        if energy_shift > 0:
            es = energy_shift
            print(f"add {es} [eV] energy correction to each structure!!!")
            for atoms in selected_frames:
                new_energy = atoms.get_potential_energy() + es
                new_forces = atoms.get_forces(apply_constraint=False).copy()
                calc = SinglePointCalculator(
                    atoms, energy=new_energy, forces=new_forces)
                atoms.calc = calc

        print("Writing structure file... ")
        selected_frames.sort(key=lambda a:a.get_potential_energy()) # sort energy
        write(self.prefix / ("miaow-sel.xyz"), selected_frames)
        print("")

        return len(all_indices), nconverged

    @staticmethod
    def __binned_statistic(x, values, func, nbins, limits):
        '''The usage is nearly the same as scipy.stats.binned_statistic'''
        from scipy.sparse import csr_matrix

        N = len(values)
        r0, r1 = limits

        digitized = (float(nbins)/(r1 - r0)*(x - r0)).astype(int)
        S = csr_matrix((values, [digitized, np.arange(N)]), shape=(nbins, N))

        return [func(group) for group in np.split(S.data, S.indptr[1:-1])]
    
    def __run_scalculation(self):
        """ run calculation to obtain dataset statistics (rmse)
        """

        return
    
    def show_statistics(self):
        """
        """
        if not self.is_single:
            # analyse all results
            #energies, forces, en_rmse, force_rmse
            title = ("#{:<23s}  {:<12s}  {:<12s}  {:<12s}  {:<12s}  ").format(
                "SysName", "natoms", "nframes", "dE_RMSE", "dE_Std"
            )
            for x in self.global_type_list:
                title += "{:<12s}  {:<12s}  ".format(x+"_dF_RMSE", x+"_dF_Std")
            title += "\n"
            with open("all_rmse.dat", "w") as fopen:
                fopen.write(title)

        nframes_dict = {}
        rmse_results = {}
        all_energies = None
        all_forces = None
        for sys_name in self.systems.keys():
            print(f"\n\n===== System {sys_name} =====")
            # update chemical info to this system
            # TODO: change this part to the DataSystem
            chemical_formula = "".join(
                [k+str(v) for k, v in self.systems[sys_name]["composition"].items()]
            )
            self.type_list, self.type_numbers = self.__parse_type_list(
                chemical_formula
            )
            # update a few outputs path based on system
            skipped = False
            self.prefix = Path.cwd() / sys_name
            if self.prefix.exists():
                print(f"result dir {self.prefix} exists, skip compression...")
                # check MLP data
                print("Use saved property data...")
                existed_data = self.prefix / (sys_name + "-MLP.xyz")
                # TODO: check
                if existed_data.exists():
                    new_frames = read(existed_data, ":")
                    calc_name = self.calc.name.lower()

                    ref_energies, ref_forces = xyz2results(new_frames, calc=None)
                    natoms_array = np.array([len(x) for x in new_frames])

                    mlp_energies = [a.info[calc_name+"_energy"] for a in new_frames]
                    mlp_forces = merge_predicted_forces(new_frames, calc_name) 

                    energies = np.array([ref_energies, mlp_energies])
                    energies = np.array([ref_energies, mlp_energies]) / natoms_array
                    forces = [ref_forces, mlp_forces]
                    
                    en_rmse = {}
                    en_rmse["energy"] = rms_dict(energies[0], energies[1])
                    force_rmse = {}
                    for key in forces[0].keys():
                        force_rmse[key] = rms_dict(
                            forces[0][key], forces[1][key]
                        )
                else:
                    # read frames
                    sys_frames = self.read_single_system(self.main_dir / sys_name, self.pattern)
                    if sys_frames:
                        energies, forces, en_rmse, force_rmse = self.show_single_statistics(
                            sys_name, sys_frames
                        )
                    else:
                        skipped = True
            else:
                self.prefix.mkdir()
                # read frames
                sys_frames = self.read_single_system(self.main_dir / sys_name, self.pattern)
                if sys_frames:
                    energies, forces, en_rmse, force_rmse = self.show_single_statistics(
                        sys_name, sys_frames
                    )
                else:
                    skipped = True

            # --- add info ---
            if not self.is_single and not skipped:
                # add system-wise rmse results
                nframes_dict[sys_name] = energies.shape[1]
                rmse_results[sys_name] = (en_rmse, force_rmse)
                # merge energy and forces
                if all_energies is None:
                    all_energies = energies
                else:
                    all_energies = np.hstack([all_energies, energies])
                print("energy shape: ", all_energies.shape)
                if all_forces is None:
                    all_forces = forces
                    for key in forces[0].keys():
                        print(f"ref {key} force length: ", len(all_forces[0][key]))
                    for key in forces[1].keys():
                        print(f"mlp {key} force length: ", len(all_forces[1][key]))
                else:
                    # ref
                    for key in forces[0].keys():
                        if key in all_forces[0].keys():
                            all_forces[0][key].extend(forces[0][key])
                        else:
                            all_forces[0][key] = forces[0][key]
                        print(f"ref {key} force length: ", len(all_forces[0][key]))
                    # mlp
                    for key in forces[1].keys():
                        if key in all_forces[1].keys():
                            all_forces[1][key].extend(forces[1][key])
                        else:
                            all_forces[1][key] = forces[1][key]
                        print(f"mlp {key} force length: ", len(all_forces[1][key]))

                natoms = sum(self.systems[sys_name]["composition"].values())
                content = "{:<24s}  {:<12d}  {:<12d}  ".format(sys_name, natoms, nframes_dict[sys_name])
                content += ("{:<12.4f}  "*2).format(en_rmse["energy"]["rmse"], en_rmse["energy"]["std"])
                for x in self.global_type_list:
                    force_rmse_x = force_rmse.get(x, {})
                    x_rmse = force_rmse_x.get("rmse", np.NaN)
                    x_std = force_rmse_x.get("std", np.NaN)
                    content += ("{:<12.4f}  "*2).format(x_rmse, x_std)
                content += "\n"

                with open("all_rmse.dat", "a") as fopen:
                    fopen.write(content)

        if not self.is_single and all_energies is not None:
            # all 
            en_rmse, force_rmse = self.__plot_comparasion(
                "ALL", all_energies, all_forces, "ALL-cmp.png"
            )
            content = "{:<24s}  {:<12d}  {:<12d}  ".format("ALL", 0, sum(nframes_dict.values()))
            content += ("{:<12.4f}  "*2).format(en_rmse["energy"]["rmse"], en_rmse["energy"]["std"])
            for x in self.global_type_list:
                force_rmse_x = force_rmse.get(x, {})
                x_rmse = force_rmse_x.get("rmse", np.NaN)
                x_std = force_rmse_x.get("std", np.NaN)
                content += ("{:<12.4f}  "*2).format(x_rmse, x_std)
            content += "\n"

            with open("all_rmse.dat", "a") as fopen:
                fopen.write(content)

        return

    def show_single_statistics(
        self, 
        sys_name,
        frames
    ):
        """ parameters
            fmax: 0.05
        """
        print(f"----- {sys_name} statistics -----")
        # results
        res_dir = self.prefix

        # check converged structures
        # NOTE: ignore converged structures now
        #converged_frames, converged_indices = self.find_converged(self.name, frames, self.prefix)
        #nconverged = len(converged_indices)
        #print("Number of converged: ", nconverged)

        #if len(converged_frames) > 0:
        #    converged_energies = [a.get_potential_energy() for a in converged_frames]
        #    converged_frames.sort(key=lambda a:a.get_potential_energy())
        #    write(res_dir / "local_minimum.xyz", converged_frames)

        # use loaded frames
        nframes = len(frames)
        ref_energies, ref_forces = xyz2results(frames, calc=None)
        natoms_array = np.array([len(x) for x in frames])

        if self.calc is not None:
            st = time.time()
            # TODO: compare with MLP
            existed_data = res_dir / (sys_name + "-MLP.xyz")
        
            # use loaded frames
            calc_name = self.calc.name.lower()
            print("Calculating with MLP...")
            new_frames = append_predictions(frames, calc=self.calc)
            #mlp_energies, mlp_forces = xyz2results(frames, calc=calc)
            mlp_energies = [a.info[calc_name+"_energy"] for a in new_frames]
            mlp_forces = merge_predicted_forces(new_frames, calc_name) 
            energies = np.array([ref_energies, mlp_energies])

            # save data
            write(existed_data, new_frames)

            forces = np.array([ref_forces, mlp_forces])
            energies = np.array([ref_energies, mlp_energies]) / natoms_array

            et = time.time()
            print("calculation time: ", et-st)

            en_rmse, force_rmse = self.__plot_comparasion(calc_name, energies, forces, res_dir / (sys_name + "-cmp.png"))

        # plot stat
        fig, axarr = plt.subplots(nrows=1, ncols=2, figsize=(16, 12))
        axarr = axarr.flatten()

        plt.suptitle("Dataset Overview")

        plot_hist(axarr[0], np.array(ref_energies).flatten(),
                  "Energy [eV]", "Number of Frames")
        for x in self.type_list:
            plot_hist(
                axarr[1], np.array(ref_forces[x]).flatten(
                ), "Force [eV/AA]", "Number of Frames",
                label=x
            )
        axarr[1].legend()

        plt.savefig(res_dir / (sys_name + "-stat.png"))
        
        # save rmse dat
        self.__write_rmse_results(
            sys_name, natoms_array[0], nframes, en_rmse, force_rmse
        )

        return energies, forces, en_rmse, force_rmse

    @staticmethod
    def test_uncertainty_consistent(frames, calc, saved_figure=None):
        # use loaded frames
        ref_energies, ref_forces = xyz2results(frames, calc=None)
        natoms_array = np.array([len(x) for x in frames])
        ref_energies = np.array(ref_energies)

        mlp_energies, mlp_forces, mlp_svars = xyz2results(
            frames, calc=calc, other_props=["en_stdvar", "force_stdvar"]
        )
        mlp_energies = np.array(mlp_energies)
        en_stdvar = np.array(mlp_svars["en_stdvar"])

        # save energy data
        ener_data = np.vstack([natoms_array, ref_energies, mlp_energies, en_stdvar])
        np.savetxt("en_devi.dat", ener_data.T)

        # save forces data
        forces_O = np.vstack(
            [np.array(ref_forces["O"]), np.array(mlp_forces["O"])]
        )
        np.savetxt("forces_O.dat", forces_O.T)
        forces_Pt = np.vstack(
            [np.array(ref_forces["Pt"]), np.array(mlp_forces["Pt"])]
        )
        np.savetxt("forces_Pt.dat", forces_Pt.T)

        # save data
        #with open("./ref_data.json", "w") as fopen:
        #    json.dump(fopen, ref_energies)

        # plot data
        fig, axarr = plt.subplots(
            nrows=2, ncols=2,
            gridspec_kw={'hspace': 0.3}, figsize=(16, 16)
        )
        axarr = axarr.flat[:]

        fig.suptitle("Committee Model Uncertainty Estimation")

        ax = axarr[0]
        parity_plot_dict(
            {'energy': np.array(ref_energies)/natoms_array},
            {'energy': np.array(mlp_energies)/natoms_array},
            ax,
            {
                "xlabel": "DFT [eV]",
                "ylabel": "%s [eV]" % calc.name.upper(),
                "title": "(a) Energy Comparasion"
            }
        )

        ax = axarr[1]
        true_deviations = np.fabs(
            np.array(ref_energies) - np.array(mlp_energies)) / natoms_array
        predicted_deviations = mlp_svars["en_stdvar"] / natoms_array
        parity_plot_dict(
            {"deviation": predicted_deviations},
            {"deviation": true_deviations},  # absolute?
            ax,
            {
                "xlabel": "%s Uncertainty [eV]" % calc.name.upper(),
                "ylabel": "True Error [eV]",
                "title": "(b) Energy Uncertainty"
            },
            [1.0, 10.0],
            write_info=False
        )

        # TODO: force deviation
        ax = axarr[2]
        parity_plot_dict(
            {"Pt": ref_forces["Pt"]},
            {"Pt": mlp_forces["Pt"]},
            ax,
            {
                "xlabel": "DFT [eV/Å]",
                "ylabel": "%s [eV/Å]" % calc.name.upper(),
                "title": "(c) Platinum Forces"
            }
        )

        ax = axarr[3]
        parity_plot_dict(
            {"O": ref_forces["O"]},
            {"O": mlp_forces["O"]},
            ax,
            {
                "xlabel": "DFT [eV/Å]",
                "ylabel": "%s [eV/Å]" % calc.name.upper(),
                "title": "(d) Oxygen Forces"
            }
        )

        plt.savefig(saved_figure)

        return

    @staticmethod
    def test_uncertainty_evolution(frames, calc, saved_file=None):
        def recalc_atoms(calc, atoms):
            # can't pickle torch or tf
            calc.reset()
            calc.calc_uncertainty = True
            atoms.calc = calc
            dummy = atoms.get_forces()
            singlepoint = SinglePointCalculator(
                atoms, energy=atoms.calc.results["energy"],
                forces=atoms.calc.results["forces"],
            )
            atoms.info["en_stdvar"] = atoms.calc.results["en_stdvar"]
            atoms.calc = singlepoint
            return atoms
        # overwrite info in frames
        print("Test Uncertainty Evolution...")
        saved_file = Path(saved_file)
        if saved_file.exists():
            frames = read(saved_file, ":")
        else:
            # frames = Parallel(n_jobs=4)(delayed(recalc_atoms)(calc, atoms) for atoms in frames)
            for atoms in tqdm(frames):
                calc.reset()
                calc.calc_uncertainty = True
                atoms.calc = calc
                dummy = atoms.get_forces()
                singlepoint = SinglePointCalculator(
                    atoms, energy=atoms.calc.results["energy"],
                    forces=atoms.calc.results["forces"],
                )
                atoms.info["en_stdvar"] = atoms.calc.results["en_stdvar"]
                atoms.calc = singlepoint
            write(saved_file, frames)

        # write to file
        content = "#{}  {}  {}  {}\n".format(
            "step   ", "natoms  ", "energy  ", "deviation")
        for i, atoms in enumerate(frames):
            content += ("{:8d}  "*2+"{:8.4f}  "*2+"\n").format(
                i, len(atoms), atoms.get_potential_energy(
                ), atoms.info["en_stdvar"]
            )
        with open("evolution.dat", "w") as fopen:
            fopen.write(content)

        fig, axarr = plt.subplots(
            nrows=2, ncols=1,
            gridspec_kw={'hspace': 0.3}, figsize=(16, 12)
        )
        axarr = axarr.flatten()
        plt.suptitle("GCMC Evolution")

        ax = axarr[0]
        # element_numbers = Counter()
        natoms_array = np.array([len(a) for a in frames])
        # oxygen_array = np.array([Counter(a.get_chemical_symbols())["O"] for a in frames])
        steps = range(len(frames))
        energies = np.array([a.get_potential_energy()
                            for a in frames]) / natoms_array
        en_stdvars = np.array([a.info["en_stdvar"]
                              for a in frames]) / natoms_array

        ax.set_xlabel("MC Step")
        ax.set_ylabel("Energy [eV]")

        ax.plot(steps, energies, label="energy per atom")
        apex = 10.
        ax.fill_between(
            steps, energies-apex*en_stdvars, energies+apex*en_stdvars, alpha=0.2,
            label="10 times deviation per atom"
        )
        for i in range(1, 5001, 500):
            ax.text(
                steps[i], energies[i]+apex *
                en_stdvars[i], "%.4f" % en_stdvars[i],
                fontsize=16
            )

        ax.legend()

        # coverage and number of oxygen atoms
        ax = axarr[1]

        nsurfatoms = int(len(frames[0])/6)
        print("Number of surface atoms: ", nsurfatoms)
        noxygen_array = np.array(
            [Counter(a.get_chemical_symbols())["O"] for a in frames])
        coverages = noxygen_array / nsurfatoms
        ax.plot(steps, coverages, color="r", label="coverage of O")

        ax.set_xlabel("MC Step")
        ax.set_ylabel("Coverage [ML]")

        ax.legend()

        plt.savefig("m-stdvar.png")

        return
    
    def __write_rmse_results(
        self, sys_name, natoms, nframes,
        en_rmse, force_rmse
    ):
        """"""
        # save rmse dat
        title = ("#{:<23s}  "+"{:<12s}  "*4).format(
            "SysName", "natoms", "nframes", "dE_RMSE", "dE_Std"
        )
        for x in self.type_list:
            title += "{:<12s}  {:<12s}  ".format(x+"_dF_RMSE", x+"_dF_Std")
        title += "\n"

        content = "{:<24s}  {:<12d}  {:<12d}  ".format(sys_name, natoms, nframes)
        content += ("{:<12.4f}  "*2).format(en_rmse["energy"]["rmse"], en_rmse["energy"]["std"])
        for x in self.type_list:
            content += ("{:<12.4f}  "*2).format(force_rmse[x]["rmse"], force_rmse[x]["std"])
        content += "\n"

        with open(self.prefix / "rmse.dat", "w") as fopen:
            fopen.write(title+content)
        
        print(title+content)

        return

    @staticmethod
    def __plot_comparasion(calc_name, energies, forces, saved_figure):
        """"""
        fig, axarr = plt.subplots(
            nrows=1, ncols=2,
            gridspec_kw={'hspace': 0.3}, figsize=(16, 12)
        )
        axarr = axarr.flatten()

        _, nframes = energies.shape
        plt.suptitle(f"Number of frames {nframes}")

        ax = axarr[0]
        en_rmse_results = parity_plot_dict(
            {"energy": energies[0, :]}, {"energy": energies[1, :]}, 
            ax, 
            {
                "xlabel": "DFT [eV]", "ylabel": "%s [eV]" % calc_name, "title": "Energy"
            }
        )

        ax = axarr[1]
        for_rmse_results = parity_plot_dict(
            forces[0], forces[1],
            ax,
            {
                "xlabel": "DFT [eV/AA]",
                "ylabel": "%s [eV/AA]" % calc_name,
                "title": "Forces"
            }
        )

        plt.savefig(saved_figure)

        return en_rmse_results, for_rmse_results


if __name__ == "__main__":
    # print("See usage in utils.anal_data")
    # test DatasetSystems
    ds = DatasetSystems("/mnt/scratch2/users/40247882/PtOx-dataset/systems.yaml")
    print(ds)
    ds.read_all_systems("/users/40247882/scratch2/oxides/eann-main/Latest-Compressed")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import abc
import copy
import logging
import pathlib

from . import config
from . import registers
from . import ComputerVariable, DriverBasedWorker, SingleWorker

from ..core.node import AbstractNode


class AbstractExpedition(AbstractNode):

    #: Name of the expedition.
    name: str = "expedition"

    @abc.abstractmethod
    def read_convergence(self):

        return

    @abc.abstractmethod
    def get_workers(self):

        return

    def run(self, *args, **kwargs) -> None:
        """"""
        # - some imported packages change `logging.basicConfig`
        #   and accidently add a StreamHandler to logging.root
        #   so remove it...
        for h in logging.root.handlers:
            if isinstance(h, logging.StreamHandler) and not isinstance(
                h, logging.FileHandler
            ):
                logging.root.removeHandler(h)

        assert self.worker is not None, "MC has not set its worker properly."

        return

    def register_worker(self, worker: dict, *args, **kwargs):
        """"""
        if isinstance(worker, dict):
            worker_params = copy.deepcopy(worker)
            worker = registers.create(
                "variable", "computer", convert_name=True, **worker_params
            ).value[0]
        elif isinstance(worker, list):  # assume it is from a computervariable
            worker = worker[0]
        elif isinstance(worker, ComputerVariable):
            worker = worker.value[0]
        elif isinstance(worker, DriverBasedWorker) or isinstance(worker, SingleWorker):
            worker = worker
        else:
            raise RuntimeError(f"Unknown worker type {worker}")

        self.worker = worker

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*

import time
from typing import List

import numpy as np

import ase
from ase import Atoms
from ase.io import read, write
from ase.collections import g2
from ase.build import molecule

from .builder import StructureBuilder


def compute_number_ideal_gas_molecules(pressure: float, volume: float, temperature: float):
    """Compute number of gas molecules by pV = nRT.

    1 eV/Ang3 = 160.21766208 GPa
    1 eV/Ang3 = 1581225.3844559586 atm
    1 atm = 1.01325 bar = 0.000101325 GPa
    1 atm = 6.32420912180121e-07 eV/Ang3
    1 Pa = 6.241509125883257e-12 eV/Ang3

    """
    #R = 8.314
    #nspecies = (pressure*volume) / (R*temperature) * 6.0221367e23
    kB = 1.38 * 1e-23 # J/K
    nmolecules = (pressure*volume) / (kB*temperature)

    return nmolecules


def build_species(species):
    # - build adsorbate
    atoms = None
    if species in ase.data.chemical_symbols:
        atoms = Atoms(species, positions=[[0.,0.,0.]])
    elif species in g2.names:
        atoms = molecule(species)
        #print("molecule: ", atoms.positions)
        #exit()
    else:
        raise ValueError(f"Cant create species {species}")

    return atoms


class MoleculeBuilder(StructureBuilder):

    name: str = "molecule"

    #: Default box for molecule to stay in. 
    default_box = np.eye(3)*20.

    #: Whether centre the molecule.
    recentre: bool = True

    def __init__(self, name: str=None, filename: str=None, box=None, recentre=True, directory="./", *args, **kwargs):
        """"""
        super().__init__(directory=directory, *args, **kwargs)

        if (name is None and filename is None) or (name is not None and filename is not None):
            raise RuntimeError(f"Either name or filname should be set for {self.name}.")
        if name is not None:
            self.molecule_name, self.func = name, self._build
        if filename is not None:
            self.filename, self.func = filename, self._read
        
        if box is not None:
            self.default_box = box
        self.recentre = recentre

        return
    
    def run(self, *args, **kwargs) -> List[Atoms]:
        """"""
        super().run(*args, **kwargs)

        atoms = self.func()

        # - check box
        #   NOTE: for molecule without cell,
        #         atoms.get_cell() returns [0.,0.,0.] while atoms.get_cell(complete=True)
        #         returns [1.,1.,1.]
        cell = atoms.get_cell()
        if np.sum(np.fabs(cell)) < 1e-8:
            # The box is too small.
            atoms.set_cell(self.default_box)
        else:
            ...
        
        # - centre the molecule
        cell = atoms.get_cell(complete=True)
        if self.recentre:
            box_centre = np.sum(cell, axis=0)/2.
            cop = np.average(atoms.get_positions(), axis=0)
            atoms.translate(box_centre-cop)
            ...

        return [atoms]
    
    def _read(self) -> Atoms:
        """"""
        frames = read(self.filename, ":")
        nframes = len(frames)
        assert nframes == 1, "Only one structure is accepted."

        return frames[0]
    
    def _build(self) -> Atoms:
        """"""
        molecule_name = self.molecule_name
        # - build adsorbate
        atoms = None
        if molecule_name in ase.data.chemical_symbols:
            atoms = Atoms(molecule_name, positions=[[0.,0.,0.]])
        elif molecule_name in g2.names:
            atoms = molecule(molecule_name)
        else:
            raise RuntimeError(f"Cant create molecule {molecule_name}")

        return atoms


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
from typing import Callable, Tuple, List

import networkx as nx

from joblib import Parallel, delayed

from ase import Atoms
from ase.io import read, write

from .. import CustomTimer
from .. import StruGraphCreator, extract_chem_envs
from .. import get_unique_environments_based_on_bonds
from .. import unpack_node_name

from .modifier import GraphModifier, DEFAULT_GRAPH_PARAMS
from ..group import create_a_group


def single_exchange_adsorbate(
    graph_params: dict,
    species: str,
    target,
    target_group,
    atoms: Atoms,
    print_func: Callable = print,
    debug_func: Callable = print,
) -> Tuple[List[Atoms], List[nx.Graph]]:
    """Exchange selected particles from the structure with target species.

    Currently, only single atom can be removed. TODO: molecule.

    Args:
        graph_params: Parameters for creating graphs.
        spec_params: Parameters for finding species to remove.

    """
    # - create graph from structure
    stru_creator = StruGraphCreator(**graph_params)

    # - check if spec_indices are all species
    natoms = len(atoms)
    group_indices = list(range(natoms))
    for command in target_group:
        curr_indices = create_a_group(atoms, command)
        group_indices = [i for i in group_indices if i in curr_indices]
    debug_func(f"group_indices to remove {group_indices}")

    # TODO: tags for molecule?
    chemical_symbols = atoms.get_chemical_symbols()
    for i in group_indices:
        if chemical_symbols[i] == species:
            break
    else:
        raise RuntimeError(f"There is no {species} to remove by target group.")

    # NOTE: compare environments of host species
    # - get chem envs
    graph = stru_creator.generate_graph(atoms, ads_indices=group_indices)
    chem_envs = extract_chem_envs(
        graph, atoms, group_indices, stru_creator.graph_radius
    )
    # print("ex chem envs: ", len(chem_envs))
    # print(chem_envs[0])
    # print(chem_envs[1])
    # for i, g in enumerate(chem_envs):
    #    print(g)
    #    #plot_graph(g, f"graph-{i}.png")
    #    for (u, d) in g.nodes.data():
    #        if d["central_ads"]:
    #            print(u, d)
    # NOTE: for single atom adsorption,
    assert len(chem_envs) == len(
        group_indices
    ), "Single atoms group into one adsorbate. Try reducing the covalent radii."
    # TODO: for molecule adsorption

    # - find unique sites to remove for this structure
    unique_indices = get_unique_environments_based_on_bonds(chem_envs)
    unique_envs = [chem_envs[i] for i in unique_indices]

    # - create sctructures
    unique_frames = []
    for g in unique_envs:
        for u, d in g.nodes.data():
            if d["central_ads"]:
                chem_sym, idx, offset = unpack_node_name(u)
                if chem_sym == species:
                    # new_atoms = make_clean_atoms(atoms)
                    new_atoms = atoms.copy()
                    new_atoms[idx].symbol = target  # TODO: a molecule?
                    unique_frames.append(new_atoms)
                    break
        else:
            # no valid adsorbate for this structure
            ...

    return unique_frames, unique_envs


class GraphExchangeModifier(GraphModifier):

    def __init__(
        self,
        species: str,
        target: str,
        target_group,
        spectators: List[str],
        substrates=None,
        graph: dict = DEFAULT_GRAPH_PARAMS,
        *args,
        **kwargs,
    ):
        """Insert an adsorbate on sites according to graph representation."""
        super().__init__(substrates=substrates, *args, **kwargs)

        self.species = species  # make this a node
        self.target = target

        self.target_group = target_group

        self.spectators = spectators
        self.graph_params = graph

        # self.check_site_unique = True
        # adsorbate_indices
        # site_radius
        # region

        return

    def _irun(self, substrates: List[Atoms]) -> List[Atoms]:
        """Exchange an adsorbate with another species."""
        self._print("---run exchange---")
        # - params for graph creator
        graph_params = copy.deepcopy(self.graph_params)
        adsorbate_elements = copy.deepcopy(self.spectators)
        graph_params.update(adsorbate_elements=adsorbate_elements)

        # - get possible sites to exchange
        with CustomTimer(name="exchange-adsorbate", func=self._print):
            ret = Parallel(n_jobs=self.njobs)(
                delayed(single_exchange_adsorbate)(
                    graph_params,
                    self.species,
                    self.target,
                    self.target_group,
                    a,
                    print_func=self._print,
                    debug_func=self._debug,
                )
                for a in substrates
            )

            ret_frames, ret_envs = [], []
            for i, (frames, envs) in enumerate(ret):
                nenvs = len(envs)
                # TODO: add info since it may be lost in atoms.copy() function
                # for a in frames:
                #    a.info["subid"] = subid
                # -- add data
                ret_envs.extend(envs)
                ret_frames.extend(frames)
                self._print(f"number of sites {nenvs} to exchange for substrate {i}.")
        # not unique across substrates
        write(self.directory / f"possible_frames.xyz", ret_frames)

        # - compare the graph of chemical environments in the structure
        #   NOTE: if Zn atoms were to exchange with Cr, the chem envs of
        #         the rest Zn atoms are used to compare the structure difference.
        #         TODO: consider Cr as well?
        created_frames = self._compare_structures(
            ret_frames, graph_params, self.target_group
        )

        return created_frames


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import os
import pathlib

from typing import Callable, List, Tuple

from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator


"""This module stores several utilities that copy structures.
"""


def read_sort(
    directory: pathlib.Path, sort_fname: str = "ase-sort.dat"
) -> Tuple[List[int], List[int]]:
    """Create the sorting and resorting list from ase-sort.dat.

    If the ase-sort.dat file does not exist, the sorting is redone.

    """
    sortfile = directory / sort_fname
    if os.path.isfile(sortfile):
        sort = []
        resort = []
        with open(sortfile, "r") as fd:
            for line in fd:
                s, rs = line.split()
                sort.append(int(s))
                resort.append(int(rs))
    else:
        # warnings.warn(UserWarning, 'no ase-sort.dat')
        raise ValueError("no ase-sort.dat")

    return sort, resort


def resort_atoms_with_spc(
    atoms_sorted: Atoms,
    resort: List[int],
    calc_name: str,
    properties=["energy", "forces", "free_energy", "stress"],
    print_func: Callable = print,
    debug_func: Callable = print,
) -> Atoms:
    """Create a spc to store calc results since some atoms may share a calculator.

    TODO: Add charges and magmoms?

    """
    _print = print_func
    _debug = debug_func

    #
    atoms = atoms_sorted.copy()[resort]

    # The original calculator must have an energy property
    results = dict(
        energy=atoms_sorted.get_potential_energy(apply_constraint=False),
    )

    try:
        free_energy = atoms_sorted.get_potential_energy(
            force_consistent=True, apply_constraint=False
        )
    except:
        free_energy = None  # results["energy"]
        _print("No free_energy property.")
    if free_energy is not None:
        results["free_energy"] = free_energy

    try:
        forces = atoms_sorted.get_forces(apply_constraint=False)[resort]
    except:
        forces = None
        _print("No forces property.")
    if forces is not None:
        results["forces"] = forces

    try:
        stress = atoms_sorted.get_stress()
    except:
        stress = None
        _print("No stress property.")
    if stress is not None:
        results["stress"] = stress

    calc = SinglePointCalculator(atoms, **results)
    calc.name = calc_name
    atoms.calc = calc

    return atoms


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itertools

from typing import Optional

import numpy as np

from ..data.array import AtomsNDArray
from .selector import AbstractSelector


class RandomSelector(AbstractSelector):

    name = "random"

    default_parameters = dict(number=[4, 0.2])

    """"""

    def __init__(
        self, directory="./", axis: Optional[int] = None, *args, **kwargs
    ) -> None:
        """"""
        super().__init__(directory, axis, *args, **kwargs)

        return

    def _mark_structures(self, data: AtomsNDArray, *args, **kwargs) -> None:
        """"""
        marker_groups = self.group_structures_by_axis(data, self.axis)

        selected_markers = []
        for grp_name, markers in marker_groups.items():
            num_markers = len(markers)
            num_selected = self._parse_selection_number(num_markers)
            if num_selected > 0:
                curr_selected_markers = self.rng.choice(
                    markers, size=num_selected, replace=False
                )
                selected_markers.extend(curr_selected_markers)
            else:
                ...

        data.markers = np.array(selected_markers)

        return

    @staticmethod
    def group_structures_by_axis(data: AtomsNDArray, axis: Optional[int] = None):
        # - group markers
        if axis is not None:
            ndim = len(data.shape)
            if axis < -ndim or axis > ndim:
                raise IndexError(f"axis {axis} is out of dimension {ndim}.")
            if axis < 0:
                axis = ndim + axis

            marker_groups = {}
            for k, v in itertools.groupby(data.markers, key=lambda x: x[axis]):
                if k in marker_groups:
                    marker_groups[k].extend(list(v))
                else:
                    marker_groups[k] = list(v)
        else:
            marker_groups = dict(all=data.markers)

        return marker_groups


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
import dataclasses

from typing import List

import numpy as np

from ase import Atoms
from ase import units
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.neighborlist import NeighborList, natural_cutoffs

from .. import convert_string_to_atoms
from .exchange import BasicExchangeOperator


@dataclasses.dataclass
class ElementaryReaction:

    #: Particles
    particles: List[str]

    #: The standard chemical potentials.
    chempot_0: List[float]

    #: Chemical coefficients.
    coefficients: List[int]

    def __post_init__(
        self,
    ):
        """"""
        self.number_net_change = np.sum(self.coefficients)
        self.chempot_change = np.sum(
            [c * mu for c, mu in zip(self.coefficients, self.chempot_0)]
        )

        equation_parts = [[], []]
        self.reactants, self.products = [], []
        for c, p in zip(self.coefficients, self.particles):
            if c < 0:
                equation_parts[0].append(f"{-1*c}{p}")
                self.reactants.append(p)
            else:
                equation_parts[1].append(f"{c}{p}")
                self.products.append(p)
        self.chemical_equation = (
            "+".join(equation_parts[0]) + "->" + "+".join(equation_parts[1])
        )

        return

    def compute_equilibrium_constant(self, temperature: float):
        """Compute the ideal-gas equilibrium constant."""
        kBT_eV = units.kB * temperature
        beta = 1.0 / kBT_eV

        # -
        self.k = np.exp(-beta * self.chempot_change)

        # - add PV term
        p_0 = 1e5  # pascal
        ang3tom3 = 1e-30
        pv_factor = (
            beta * p_0 * ang3tom3 * units.kJ / 1000.0
        ) ** self.number_net_change

        self.k2 = pv_factor * self.k

        return


class ReactOperator(BasicExchangeOperator):

    name: str = "react"

    def __init__(
        self,
        reaction,
        region,
        temperature,
        pressure: float = 1.0,
        use_bias: bool = True,
        *args,
        **kwargs,
    ):
        """"""
        super().__init__(
            region=region, temperature=temperature, pressure=pressure, *args, **kwargs
        )

        # - parse reaction
        self.reaction = ElementaryReaction(**reaction)
        self.reaction.compute_equilibrium_constant(temperature=temperature)

        # -
        self.use_bias = use_bias

        return

    def run(self, atoms: Atoms, rng=np.random.Generator(np.random.PCG64())) -> Atoms:
        """"""
        super().run(atoms)
        self._extra_info = "-"

        # -- volume?
        if self.use_bias:
            # --- determine on-the-fly
            acc_volume = self.region.get_empty_volume(atoms)
        else:
            # --- get normal region
            acc_volume = self.region.get_volume()
        self._curr_volume = acc_volume

        # -- TODO: find possible reaction sites...
        self._print(f"reaction: {self.reaction.chemical_equation}")

        # --
        reactant_numbers = [
            len(self._curr_tags_dict.get(r, [])) for r in self.reaction.reactants
        ]
        has_reactants = all([r_n > 0 for r_n in reactant_numbers])

        product_numbers = [
            len(self._curr_tags_dict.get(p, [])) for p in self.reaction.products
        ]
        has_products = all([p_n > 0 for p_n in product_numbers])

        self._curr_particle_numbers = reactant_numbers + product_numbers

        if not has_reactants and not has_products:
            self._print("No reactants and products.")
            self._curr_operation = "skipped"
            curr_atoms = None
        elif has_reactants and not has_products:
            self._print("Only reactants.")
            self._print("...forward...")
            self._curr_operation = "forward"
            curr_atoms = self._forward_reaction(atoms, rng)
        elif not has_reactants and has_products:
            self._print("Only products.")
            self._print("...reverse...")
            self._curr_operation = "reverse"
            curr_atoms = self._reverse_reaction(atoms, rng)
        else:
            rn_rxn = rng.uniform()
            if rn_rxn < 0.5:
                self._print("...forward...")
                self._curr_operation = "forward"
                curr_atoms = self._forward_reaction(atoms, rng)
            else:
                self._print("...reverse...")
                self._curr_operation = "reverse"
                curr_atoms = self._reverse_reaction(atoms, rng)

        self._extra_info = (
            f"{self._curr_operation.capitalize()}_{self.reaction.chemical_equation}"
        )

        return curr_atoms

    def _compute_prefactor(self, particle_numbers: List[int], xi: int):
        """Compute the prefactor of the acceptance probability.

        Args:
            particle_numbers: Particle numbers.
            xi: Reaction direction 1 or -1.

        """
        # -
        particle_numbers = np.array(particle_numbers)
        number_changes = np.array(self.reaction.coefficients) * xi

        fac = 1.0
        for n, c in zip(particle_numbers, number_changes):
            fac *= np.math.factorial(n) / np.math.factorial(n + c)

        # -
        prefactor = (
            (self.reaction.k2**xi)
            * (self._curr_volume ** (self.reaction.number_net_change * xi))
            * fac
        )

        return prefactor

    def _create_a_particle(self, atoms: Atoms, species: str, rng) -> Atoms:
        """"""
        # -
        particle = convert_string_to_atoms(species)

        # - add velocity in case the mixed MC/MD is performed
        MaxwellBoltzmannDistribution(particle, temperature_K=self.temperature, rng=rng)

        # - add tag
        used_tags = set(atoms.get_tags().tolist())
        new_tag = 0
        while new_tag in used_tags:
            # NOTE: np.random only has randint
            new_tag = rng.integers(self.MIN_RANDOM_TAG, self.MAX_RANDOM_TAG)
        new_tag = int(new_tag)
        # NOTE: ase accepts int or list as tags
        particle.set_tags(new_tag)

        return particle

    def _forward_reaction(
        self, atoms_: Atoms, rng=np.random.Generator(np.random.PCG64())
    ) -> Atoms:
        """Perform a forward reaction."""
        self._curr_prefactor = self._compute_prefactor(
            self._curr_particle_numbers, xi=1
        )

        atoms = atoms_.copy()

        # - remove reactant and select reactive site
        site_positions = []
        for r in self.reaction.reactants:
            reax_indices = self._select_species(atoms, [r], rng)
            selected_species = atoms[reax_indices]
            site_positions.append(np.mean(selected_species.positions, axis=0))
            del atoms[reax_indices]
            self._print(f"remove {r} {reax_indices}")

        # -- default uses the first reactant's position
        site_position = site_positions[0]

        # - insert the product to the site position
        # TODO: we only support one product here...
        product = self.reaction.products[0]
        particle = self._create_a_particle(atoms, product, rng)
        new_tag = int(particle.get_tags()[0])
        self._print(
            f"product {particle.get_chemical_formula()} tag: {particle.get_tags()}"
        )
        atoms.extend(particle)

        particle_indices = [i for i, t in enumerate(atoms.get_tags()) if t == new_tag]
        species = atoms[particle_indices]

        new_vec = site_position - np.mean(species.positions, axis=0)
        species.translate(new_vec)

        original_positions = copy.deepcopy(species.get_positions())

        # - adjust product configuration
        nl = NeighborList(
            self.covalent_max * np.array(natural_cutoffs(atoms)),
            skin=0.0,
            self_interaction=False,
            bothways=True,
        )
        cell = atoms.get_cell(complete=True)

        for i in range(self.MAX_RANDOM_ATTEMPTS):
            species_ = self._rotate_species(species, rng=rng)
            atoms.positions[particle_indices] = copy.deepcopy(species_.positions)
            if not self.check_overlap_neighbour(nl, atoms, cell, particle_indices):
                self._print(f"succeed to insert after {i+1} attempts...")
                break
            atoms.positions[particle_indices] = original_positions
        else:
            del atoms[particle_indices]
            atoms = None

        return atoms

    def _reverse_reaction(
        self, atoms_: Atoms, rng=np.random.Generator(np.random.PCG64())
    ) -> Atoms:
        """Perform a reverse reaction."""
        self._curr_prefactor = self._compute_prefactor(
            self._curr_particle_numbers, xi=-1
        )

        # - remove products
        atoms = atoms_
        for p in self.reaction.products:
            atoms = self._remove(atoms, species=p, rng=rng)

        # - insert reactants
        for r in self.reaction.reactants:
            atoms = self._insert(atoms, species=r, rng=rng)
            if atoms is None:
                break

        return atoms

    def metropolis(self, prev_ene: float, curr_ene: float, rng=np.random) -> bool:
        """Acceptance probability.

        We have equations as
        P = (beta*P0*V)^(nu*xi)*K^xi*PROD_i^s(N0!/(N0+nu_i*xi)!)*exp(-beta*dU)
        K = \exp(-\sum_i^s{nu_i*\mu_i^0}*beta)

        nu - net number change in particles
        xi - reaction direction

        """
        # - acceptance ratio
        kBT_eV = units.kB * self.temperature
        beta = 1.0 / kBT_eV  # 1/(kb*T), eV

        ene_diff = curr_ene - prev_ene
        ene_term = np.exp(-beta * ene_diff)

        acc_ratio = np.min([1.0, self._curr_prefactor * ene_term])

        content = f"\nVolume {self._curr_volume:>.4f}\n"
        content += f"Prefactor {self._curr_prefactor}\n"
        content += f"Energy Difference {ene_diff:>.4f} [eV]  Exp {ene_term}\n"
        content += f"Accept Ratio {acc_ratio:>.4f}\n"
        for l in content.split("\n"):
            self._print(l)

        # -
        rn_rxn = rng.uniform()
        self._print(f"{self.__class__.__name__} Probability %.4f" % rn_rxn)

        # - reset stored temp data
        self._curr_operation = None
        self._curr_tags_dict = None
        self._curr_volume = None
        self._curr_particle_numbers = None
        self._curr_prefactor = None

        return rn_rxn < acc_ratio

    def __repr__(self) -> str:
        """"""
        content = f"@Modifier {self.__class__.__name__}\n"
        content += (
            f"temperature {self.temperature} [K] pressure {self.pressure} [bar]\n"
        )
        content += "covalent ratio: \n"
        content += f"  min: {self.covalent_min} max: {self.covalent_max}\n"
        content += f"reaction: "
        content += f"  {self.reaction.chemical_equation}\n"
        content += f"  within the region {self.region}\n"

        return content

    def as_dict(self) -> dict:
        """"""
        params = super().as_dict()

        return params


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
import time
import logging
import pathlib

from typing import Union, List, NoReturn
from collections import Counter


from ase import Atoms
from ase.io import read, write
from ase.formula import Formula


class StructureSystem():

    """A system contains structures from either a folder or a database.
    """

    name = None

    desc = ""
    chemical_formula:Formula = None
    substrate = "" # molecule, bulk, surface, cluster
    composition:dict = {}

    frames = None

    pfunc = print
    
    #: System-dependent parameters, e.g., kpts.
    specific_params:dict = {}

    def __init__(self, origin, name, composition:dict =None, *args, **kwargs) -> None:
        """"""
        self.origin = pathlib.Path(origin) # TODO: either a folder or a database
        self.name = name

        self.desc, self.chemical_formula, self.substrate = self._parse_name_components()
        composition_ = dict(Counter(self.chemical_formula))
        if composition is not None:
            assert composition == composition_, "Input composition is inconsistent with name."
        self.composition = composition_

        self.specific_params = kwargs

        return

    def _parse_name_components(self):
        """System name should be one of three formats.

        1. simple chemical formula. 2. custom name plus chemical formula.
        3. custom name, chemical formula and system type.

        """
        desc, substrate = "", ""
        name_parts = self.name.split("-")
        if len(name_parts) == 1:
            chemical_formula = name_parts[0]
        elif len(name_parts) == 2:
            desc, chemical_formula = name_parts
        elif len(name_parts) == 3:
            desc, chemical_formula, substrate = name_parts
        else:
            raise ValueError("directory name must be as xxx, xxx-xxx, or xxx-xxx-xxx")
        chemical_formula = Formula(chemical_formula)

        return desc, chemical_formula, substrate
    
    def _read_structures(self) -> List[Atoms]:
        """"""
        sys_path = self.origin/self.name

        total_frames = []
        stru_files = list(sys_path.glob("*.xyz")) # TODO: support more formats
        stru_files.sort()
        for p in stru_files:
            frames = read(p, ":")
            total_frames.extend(frames)
            self.pfunc(f"  structure path: {str(p)}")
            self.pfunc(f"    number of frames: {len(frames)}")
        self.pfunc(f"  nframes: {len(total_frames)}")

        self.frames = total_frames

        return total_frames
    
    def _mask_structures(self):
        """"""

        return


class StructureDatabase:

    """
    """

    name = "StructureDatabase"

    #: Working directory.
    _directory: pathlib.Path = None

    restart = False

    def __init__(self, config_params: dict, pot_worker=None, ref_worker=None) -> None:
        """Initialise a structure database.

        Args:
            config_params: Configuration parameters.

        """
        self.directory = pathlib.Path.cwd()
        self._init_logger(self.directory)

        # params should have
        #   - database
        #   - type_list
        #   - systems
        self.database = config_params.get("database", None) # path
        assert self.database, "Database is undefined."
        self.database = pathlib.Path(self.database)

        system_dict = config_params.get("systems", {})

        # - create systems
        self.systems = []
        for name, sys_params in system_dict.items():
            sys_params = copy.deepcopy(sys_params)
            composition = sys_params.pop("composition", None)
            sys = StructureSystem(
                origin=self.database, name=name, composition=composition,
                **sys_params
            )
            sys.pfunc = self.logger.info
            self.systems.append(sys)
        assert len(self.systems) > 0, "No systems are found."

        self.pot_worker = pot_worker
        self.ref_worker = ref_worker

        return

    @property
    def directory(self) -> pathlib.Path:
        """Working directory.

        Note:
            When setting directory, info_fpath would be set as well.

        """
        return self._directory
    
    @directory.setter
    def directory(self, directory_) -> NoReturn:
        self._directory = pathlib.Path(directory_)

        return 

    def _init_logger(self, working_directory):
        """"""
        self.logger = logging.getLogger(__name__)

        log_level = logging.INFO

        self.logger.setLevel(log_level)

        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )

        working_directory = pathlib.Path(working_directory)
        log_fpath = working_directory / (self.name+".out")

        if self.restart:
            fh = logging.FileHandler(filename=log_fpath, mode="a")
        else:
            fh = logging.FileHandler(filename=log_fpath, mode="w")

        fh.setLevel(log_level)
        #fh.setFormatter(formatter)

        ch = logging.StreamHandler()
        ch.setLevel(log_level)
        #ch.setFormatter(formatter)

        #self.logger.addHandler(ch)
        #self.logger.addHandler(fh)

        # begin!
        self.logger.info(
            "\nStart at %s\n", 
            time.asctime( time.localtime(time.time()) )
        )

        return
    
    def run(self, params: dict):
        """"""
        # - summarise systems
        nframes = 0
        for sys in self.systems:
            self.logger.info(f"SystemName: {sys.name}")
            _ = sys._read_structures()
            nframes += len(sys.frames)
        self.logger.info(f"Total Number of Frames: {nframes} by {len(self.systems)} systems.")

        # - operations...
        assert len(params), "Currently, only one operation is supported."
        for op_name, op_params in params.items():
            if op_name == "select":
                self._select(op_params)
            elif op_name == "label":
                self._label(op_params)
            break

        return
    
    def _select(self, params):
        """"""
        from gdpx.selector import create_selector
        selector = create_selector(params)
        selector.logger = self.logger

        out_path = self.directory / "selected"
        if not out_path.exists():
            out_path.mkdir()

        for sys in self.systems:
            sys_sel_path = out_path/sys.name
            if not sys_sel_path.exists():
                sys_sel_path.mkdir()
            selector.directory = sys_sel_path
            selected_frames = selector.select(sys.frames)
            write(sys_sel_path/"sel.xyz", selected_frames)

        return
    
    def _label(self, params):
        """Label structures with given potential.

        This is useful when many structures need benchmark.

        """
        # - update some specific params of worker
        worker = self.pot_worker
        worker.logger = self.logger
        #worker._submit = False
        
        out_path = self.directory / "label"
        if not out_path.exists():
            out_path.mkdir()
        #else: # TODO: warning?
        #    raise FileNotFoundError("Directory label exists.")
        
        res_path = out_path / "results"
        if not res_path.exists():
            res_path.mkdir()

        # TODO: add an outlier detector? selection...
        #       sometimes expedition gives unphysical structures
        for sys in self.systems:
            # -- update basic info for the worker
            nframes_in = len(sys.frames)
            worker.directory = out_path/sys.name
            worker.batchsize = nframes_in

            # -- check outputs
            frames_out_path = res_path/(sys.name+".xyz")
            nframes_saved = 0
            if frames_out_path.exists():
                frames_saved = read(frames_out_path, ":")
                nframes_saved = len(frames_saved)
            if nframes_saved == nframes_in: # TODO: check if frames are consistent?
                self.logger.info(f"!! {sys.name} has been saved.")
                continue

            # --- update some specific params, e.g., kpts.
            for k in worker.driver.syswise_keys:
                v = sys.specific_params.get(k, None)
                if v:
                    worker.driver.init_params.update(**{k: v})
            worker.run(sys.frames)

            # --- retrieve results
            worker.inspect(resubmit=True)
            if worker.get_number_of_running_jobs() == 0:
                ret_frames = worker.retrieve() # TODO: option for re-collect results?
                nframes_ret = len(ret_frames)
                if nframes_in == nframes_ret:
                    self.logger.info(f"{sys.name} finished and writes structures.")
                    write(res_path/(sys.name+".xyz"), ret_frames)
                else:
                    self.logger.info(f"{sys.name} finishes {nframes_ret} out of {nframes_in}.")

        return
    
    def _compress(self):
        """Compress dataset based on Boltzmann distribution."""

        return



if __name__ == "__main__":
    pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*

import copy
import pathlib
import subprocess
from typing import NoReturn, Union, List

from abc import ABC, abstractmethod


class AbstractScheduler(ABC):
    """The abstract scheduler that implements common functions.

    A scheduler deals with the lifecycle of a job in the queue.

    Attributes:
    """

    #: The name of the scheduler.
    name: str = "abstract"

    #: A string starts at each option line.
    PREFIX: str = None

    #: The suffix of a job script.
    SUFFIX: str = None

    #: The first line of a script.
    SHELL: str = None

    #: The command used to submit jobs.
    SUBMIT_COMMAND: str = None

    #: The command used to check job status.
    ENQUIRE_COMMAND: str = None

    #: Default parameters.
    default_parameters: dict = {}

    #: Current stored parameters.
    parameters: dict = {}

    #: _script: The path of the job script.
    _script: Union[str, pathlib.Path] = None

    #: The job name.
    _job_name: str = "scheduler"

    #: Environment settings for a job.
    environs: str = None

    #: Custom commands for a job.
    user_commands: str = None

    #: The tags that a job may have in the queue.
    running_status: List[str] = []

    def __init__(self, *args, **kwargs):
        """Init an abstract scheduler.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        """
        # - update params
        self.environs = kwargs.pop("environs", "")
        self.user_commands = kwargs.pop("user_commands", "")

        # - make default params
        self.parameters = self._get_default_parameters()
        # parameters_ = kwargs.pop("parameters", None)
        # if parameters_:
        #    self.parameters.update(parameters_)
        self.parameters.update(kwargs)

        # - update some special keywords
        #   job_name

        return

    @property
    def script(self) -> Union[str, pathlib.Path]:
        """Store the path of the job script."""

        return self._script

    @script.setter
    def script(self, script_):
        self._script = pathlib.Path(script_)
        return

    @property
    def job_name(self) -> str:

        return self._job_name

    @job_name.setter
    @abstractmethod
    def job_name(self, job_name_: str):
        self._job_name = job_name_
        # update job name in parameters
        return

    def _get_default_parameters(self):
        return copy.deepcopy(self.default_parameters)

    def set(self, **kwargs) -> None:
        """Set parameters.

        Args:
            **kwargs: Arbitrary keyword arguments.

        """
        # changed_parameters = {}
        for key, value in kwargs.items():
            oldvalue = self.parameters.get(key)
            # if key not in self.parameters or not equal(value, oldvalue):
            #    changed_parameters[key] = value
            #    self.parameters[key] = value
            self.parameters[key] = value

        return

    def write(self) -> None:
        """Write self to the path of the job script."""
        with open(self.script, "w") as fopen:
            fopen.write(str(self))

        return

    def submit(self) -> str:
        """Submit job using specific scheduler command and return job id."""
        command = "{0} {1}".format(self.SUBMIT_COMMAND, self.script.name)
        proc = subprocess.Popen(
            command,
            shell=True,
            cwd=self.script.parent,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="utf-8",
        )
        errorcode = proc.wait(timeout=10)  # 10 seconds
        if errorcode:
            raise RuntimeError(f"Error in submitting job script {str(self.script)}")

        output = "".join(proc.stdout.readlines())
        job_id = output.strip().split()[-1]

        return job_id

    @abstractmethod
    def is_finished(self) -> bool:
        """Check whether the job is finished.

        The job is the one performed in the current job script path.

        """

        return

    def as_dict(self) -> dict:
        """"""
        sch_params = {}
        sch_params = {k: v for k, v in self.parameters.items() if v is not None}
        sch_params["environs"] = self.environs
        sch_params["backend"] = self.name

        sch_params = copy.deepcopy(sch_params)

        return sch_params


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*


import numpy as np
from scipy.spatial import distance_matrix


from ase.io import read, write


def switch_function(distances, r_cut, r_shift=0., nn=6, mm: int=None):
    """"""
    if mm is None:
        mm = nn*2

    scaled_distances = (distances - r_shift) / r_cut

    return (1 - scaled_distances**nn) / (1 - scaled_distances**mm)


def compute_coordination_number():
    """"""
    frames = read(
        "/scratch/gpfs/jx1279/copper+alumina/dptrain/r8/_explore/_mrxn/Cu13+s001p32/_sinter/3xCu13/_200ps/cand0_400K_200ps/traj.dump", 
        ":1"
    )
    nframes = len(frames)
    print(f"nframes: {nframes}")

    atoms = frames[0]

    positions = atoms.positions[-13:, :]
    print(positions)
    clusters = [
        atoms.positions[-13:, :],
        atoms.positions[-26:-13, :],
        atoms.positions[-39:-26, :],
    ]

    com_clusters = [np.mean(p, axis=0) for p in clusters]
    print(com_clusters)
    print(np.linalg.norm(com_clusters[2] - [0., 25.239, 0.] - com_clusters[1]))
    print(np.linalg.norm(com_clusters[2] - com_clusters[0]))

    def xxx(positions):
        dmat = distance_matrix(positions, positions)
        #print(dmat)

        sf = switch_function(dmat, r_cut=3.8, nn=8, mm=14)
        np.fill_diagonal(sf, 0.)
        coordination = np.sum(sf, axis=1)
        print(coordination)

        return coordination
    
    for positions in clusters:
        print(np.sum(xxx(positions)))

    return


if __name__ == "__main__":
    compute_coordination_number()
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
from typing import List

from ase import Atoms


def get_file_md5(f):
    import hashlib
    m = hashlib.md5()
    while True:
        # if not using binary
        #data = f.read(1024).encode('utf-8')
        data = f.read(1024) # read in block
        if not data:
            break
        m.update(data)
    return m.hexdigest()


def copy_minimal_frames(prev_frames: List[Atoms]):
    """Copy atoms without extra information.

    Do not copy atoms.info since it is a dict and does not maitain order.

    """
    curr_frames, curr_info = [], []
    for prev_atoms in prev_frames:
        # - copy geometry
        curr_atoms = Atoms(
            symbols=copy.deepcopy(prev_atoms.get_chemical_symbols()),
            positions=copy.deepcopy(prev_atoms.get_positions()),
            cell=copy.deepcopy(prev_atoms.get_cell(complete=True)),
            pbc=copy.deepcopy(prev_atoms.get_pbc()),
            tags = prev_atoms.get_tags() # retain this for molecules
        )
        if prev_atoms.get_kinetic_energy() > 0.: # retain this for MD
            curr_atoms.set_momenta(prev_atoms.get_momenta()) 
        curr_frames.append(curr_atoms)
        # - save info
        confid = prev_atoms.info.get("confid", -1)
        dynstep = prev_atoms.info.get("step", -1)
        prev_wdir = prev_atoms.info.get("wdir", "null")
        curr_info.append((confid,dynstep,prev_wdir))

    return curr_frames, curr_info


def read_cache_info(wdir, length=36):
    # - read extra info data
    _info_data = []
    for p in (wdir/"_data").glob("*_info.txt"):
        identifier = p.name[:length] # MD5
        with open(p, "r") as fopen:
            for line in fopen.readlines():
                if not line.startswith("#"):
                    _info_data.append(line.strip().split())
    _info_data = sorted(_info_data, key=lambda x: int(x[0]))

    return _info_data


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*

import copy
import os
import pathlib
from typing import List
import warnings

import yaml

import numpy as np

from ase.io import read, write
from ase.calculators.calculator import Calculator

from . import AbstractPotentialManager, AbstractTrainer
from . import DummyCalculator, CommitteeCalculator


class NequipTrainer(AbstractTrainer):

    name = "nequip"
    command = "nequip-train"
    freeze_command = "nequip-deploy"
    prefix = "config"

    #: Training directory.
    RUN_NAME: str = "auto"

    def __init__(
        self, config: dict, type_list: List[str], train_epochs: int=200,
        directory=".", command: str="nequip-train", freeze_command: str="nequip-deploy",
        random_seed: int=1112, *args, **kwargs
    ) -> None:
        super().__init__(
            config=config, type_list=type_list, train_epochs=train_epochs,
            directory=directory, command=command, freeze_command=freeze_command, 
            random_seed=random_seed, *args, **kwargs
        )

        # - TODO: sync type_list
        self._type_list = type_list

        return

    def _resolve_train_command(self, init_model=None) -> str:
        """"""
        train_command = self.command

        # - add options
        command = "{} {}.yaml ".format(train_command, self.name)
        if init_model is not None:
            #command += "--init-model {}".format(str(pathlib.Path(init_model).resolve()))
            raise RuntimeError(f"{self.__class__.__name__} does not support init_model.")
        command += " 2>&1 > {}.out\n".format(self.name)

        return command

    def _resolve_freeze_command(self, *args, **kwargs) -> str:
        """"""
        freeze_command = self.freeze_command

        # - add options
        command = "{} build --train-dir {} {} 2>&1 >> {}.out".format(
            freeze_command, self.RUN_NAME, self.frozen_name, self.name
        )

        return command
    
    @property
    def frozen_name(self):
        """"""
        return f"{self.name}.pth"
    
    def write_input(self, dataset, *args, **kwargs):
        """"""
        # - check dataset
        data_dirs = dataset.load()
        self._print(data_dirs)
        self._print("--- auto data reader ---")

        frames = []
        for i, curr_system in enumerate(data_dirs):
            curr_system = pathlib.Path(curr_system)
            self._print(f"System {curr_system.stem}\n")
            curr_frames = []
            subsystems = list(curr_system.glob("*.xyz"))
            subsystems.sort() # sort by alphabet
            for p in subsystems:
                # read and split dataset
                p_frames = read(p, ":")
                p_nframes = len(p_frames)
                curr_frames.extend(p_frames)
                self._print(f"  subsystem: {p.name} number {p_nframes}")
            self._print(f"  nframes {len(curr_frames)}")
            frames.extend(curr_frames)
        nframes = len(frames)
        self._print(f"nframes {nframes}")

        write(self.directory/"dataset.xyz", frames)

        n_train = int(nframes*dataset.train_ratio/dataset.batchsize)*dataset.batchsize
        n_val = nframes - n_train

        # - check train config
        # params: root, run_name, seed, dataset_seed, n_train, n_val, batch_size
        #         dataset, dataset_file_name
        train_config = copy.deepcopy(self.config)

        train_config["root"] = str(self.directory.resolve())
        train_config["run_name"] = self.RUN_NAME

        train_config["seed"] = self.rng.integers(0, 10000, dtype=int)
        train_config["dataset_seed"] = self.rng.integers(0, 10000, dtype=int)

        train_config["dataset"] = "ase"
        train_config["dataset_file_name"] = str((self.directory/"dataset.xyz").resolve())

        train_config["chemical_symbols"] = self.type_list

        train_config["batch_size"] = dataset.batchsize

        train_config["n_train"] = n_train
        train_config["n_val"] = n_val

        train_config["max_epochs"] = self.train_epochs

        with open(self.directory/f"{self.name}.yaml", "w") as fopen:
            yaml.safe_dump(train_config, fopen)

        return
    
    def read_convergence(self) -> bool:
        """"""
        converged = False
        with open(self.directory/self.RUN_NAME/"log", "rb") as fopen:
            try:  # catch OSError in case of a one line file 
                fopen.seek(-2, os.SEEK_END)
                while fopen.read(1) != b'\n':
                    fopen.seek(-2, os.SEEK_CUR)
            except OSError:
                fopen.seek(0)
            line = fopen.readline().decode()
        
        if line.strip().startswith("Cumulative wall time"):
            converged = True

        return converged


class NequipManager(AbstractPotentialManager):

    name = "nequip"
    implemented_backends = ["ase", "lammps"]

    valid_combinations = (
        ("ase", "ase"), # calculator, dynamics
        ("lammps", "ase"),
        ("lammps", "lammps")
    )
    
    def __init__(self):
        """"""
        self.committee = None

        return

    def _create_calculator(self, calc_params: dict) -> Calculator:
        """Create an ase calculator.

        Todo:
            In fact, uncertainty estimation has various backends as well.
        
        """
        calc_params = copy.deepcopy(calc_params)

        command = calc_params.pop("command", None)
        directory = calc_params.pop("directory", pathlib.Path.cwd())
        atypes = calc_params.pop("type_list", [])

        type_map = {}
        for i, a in enumerate(atypes):
            type_map[a] = i

        # --- model files
        model_ = calc_params.get("model", [])
        if not isinstance(model_, list):
            model_ = [model_]

        models = []
        for m in model_:
            m = pathlib.Path(m).resolve()
            if not m.exists():
                raise FileNotFoundError(f"Cant find model file {str(m)}")
            models.append(str(m))

        # - create specific calculator
        calc = DummyCalculator()
        if self.calc_backend == "ase":
            # return ase calculator
            try:
                import torch
                from nequip.ase import NequIPCalculator
            except:
                raise ModuleNotFoundError("Please install nequip and torch to use the ase interface.")
            calcs = []
            for m in models:
                curr_calc = NequIPCalculator.from_deployed_model(
                    model_path=m, species_to_type_name={k:k for k in atypes},
                    device=torch.device("cuda" if torch.cuda.is_available() else "cpu")
                )
                calcs.append(curr_calc)
            if len(calcs) == 1:
                calc = calcs[0]
            elif len(calcs) > 1:
                calc = CommitteeCalculator(calcs)
            else:
                ...
        elif self.calc_backend == "lammps":
            from gdpx.computation.lammps import Lammps
            flavour = calc_params.pop("flavour", "nequip") # nequip or allegro
            if models:
                pair_style = "{}".format(flavour)
                pair_coeff = "* * {}".format(models[0])
                calc = Lammps(
                    command=command, directory=directory, 
                    pair_style=pair_style, pair_coeff=pair_coeff,
                    **calc_params
                )
                # - update several params
                calc.units = "metal"
                calc.atom_style = "atomic"
                if pair_style == "nequip":
                    calc.set(**dict(newton="off"))
                elif pair_style == "allegro":
                    calc.set(**dict(newton="on"))

        return calc

    def register_calculator(self, calc_params):
        """"""
        super().register_calculator(calc_params)
        
        self.calc = self._create_calculator(calc_params)

        return


if __name__ == "__main__":
    ...